/*
    NullBot Robot Platform - Server - 2005-2011 Jason Hunt
    
    GNU GENERAL PUBLIC LICENSE Version 2, June 1991
        
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

    Contact: nulluser@gmail.com
    
    File: camera.c
*/

#include <linux/videodev.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <signal.h>

#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <pthread.h>
#include <memory.h>
#include <errno.h>

#include "utility.h"
#include "server.h"
#include "camera.h"
#include "compress.h"

#include "camera_drv.h"

#define MODULE_NAME "[Camera] "

#define FNAME_BASE  "camera_data"


int camera_server_sd;  
volatile sig_atomic_t camera_client_connected = 0;

struct video_mbuf mbuf;
struct video_mmap vid_buffer;
unsigned char *vid_map = NULL;
int vid_fd = -1;

unsigned char encode_buff[MAX_IMAGE];
unsigned char compressed[MAX_IMAGE];   // compression output buffer

unsigned int image_x = 0, image_y = 0;

pthread_t camera_t;

unsigned char *camera_buff;
unsigned int max_image = 0;


unsigned int save_image = 0;
unsigned int save_count = 0;


char camera_device[MAX_STR];

int camera_port;


/* Compute 16 bit checksum */
unsigned short get_checksum16(unsigned char *buff, unsigned int len)
{
    unsigned int i;
    short check = 0;

    for (i = 0; i < len; i++)
        check += buff[i];

    return(check);
}
/* End of get_checksum16 */



/* Send camera image */
int send_camera_image( int client_sd )
{
    unsigned int out_size;


  // log_write (MODULE_NAME "get frame \n");
    cameradrv_frame (camera_buff, max_image, &out_size);
 //  log_write (MODULE_NAME "return\n");

    short check = get_checksum16(camera_buff, out_size);
    


    // create the packet      
    char comm[] = { NC_IMAGE_SIZE,            // send compressed image
                    image_x >> 8,    // x_size high byte
                    image_x & 0xff,  // x_size low byte
                    image_y >> 8,    // y_size high byte
                    image_y & 0xff,  // y_size low byte
       
                    out_size >> 24,     // high     set the data _size
                    out_size >> 16,     //
                    out_size >> 8,      // 
                    out_size & 0xff,
                    
                    check >> 8,
                    check & 0xff
                                        
                    };   // low
 
 
 
 
 
    // Send information packet
    if (send(client_sd, comm, NC_IMAGE_SIZE_LEN, 0) == -1)
    {
        alert(MODULE_NAME "Send error\n");
        camera_client_connected = 0;
        return(1);
    }
    
    // Send compressed frame
    if (send(client_sd, camera_buff, out_size, 0) == -1)
    {
        alert(MODULE_NAME "Send error\n");
        camera_client_connected = 0;
        return(1);
    }
    

    if (save_image)
    {
        char fname[54];
        sprintf(fname, "image%010d.jpg", save_count++);

        FILE *f = fopen(FNAME_BASE, "wb");
        fwrite(camera_buff, out_size, 1, f);
        fclose(f);
        
    }
        
   // log_write("Send complete\n");
    
    return(0);
}
/* End of send camera image */


/* Het data from the client */
int camera_get_data( int sd, char *data )
{
    return(recv(sd, data, REC_BUFFER, 0));
}
/* End of get_data */


/* Catch network error signal */
void catch_camera_neterror( int sig )
{
    alert(MODULE_NAME "Unable to send data\n");

    camera_client_connected = 0;

    signal (sig, catch_camera_neterror);
}
/* End of catch network error */


/* Init the network */
int setup_camera_server( void )
{
    signal (SIGPIPE, catch_camera_neterror);
    signal (SIGURG, catch_camera_neterror);

    // setup the camera first

    camera_server_sd = socket(AF_INET, SOCK_STREAM, 0);

    if(camera_server_sd < 0) 
    {
        alert(MODULE_NAME "Unable to open socket\n");
        return(1);
    }

    struct sockaddr_in server_addr; 

    memset(&server_addr, 0, sizeof(server_addr));

    server_addr.sin_family = AF_INET;
    server_addr.sin_addr.s_addr = INADDR_ANY;
    server_addr.sin_port = htons(camera_port);
  
      // set socket to resuse
    int reuse = 1;
    int reuse_len = sizeof(int);

    if (setsockopt(camera_server_sd, SOL_SOCKET, SO_REUSEADDR, &reuse, reuse_len) == -1)
    {
        alert(MODULE_NAME "Unable to set REUSE\n");
        return(1);
    }
   
    if(bind(camera_server_sd, (struct sockaddr *) &server_addr, sizeof(server_addr)) == -1) 
    {
        alert(MODULE_NAME "Unable to bind\n");
        return(1);
    }
    
    if(listen(camera_server_sd, 1) == -1)  // listen for one only client
    {
        alert(MODULE_NAME "Unable to listen\n");
        return(1);
    }
 
    return(0);   
}
/* End of setup network */


/* Wait for a client to connect */
int camera_wait_for_client( void )
{
    alert(MODULE_NAME "Waiting for connection on port %d\n", camera_port);

    struct sockaddr_in client_addr;
    socklen_t client_len = sizeof(struct sockaddr_in);

    return(accept(camera_server_sd, (struct sockaddr *) &client_addr, &client_len));
}
/* End of wait for client */


/* Core loop for clients */
// BUG need to replace with stream processor
void camera_network_loop( int client_sd )
{
    camera_client_connected = 1;    // we are now connected

    char data[REC_BUFFER];
        
    while(!program_done && camera_client_connected) // n == 0, client disconnect
    {
        int n = recv(client_sd, data, REC_BUFFER, 0);

        if (n < 1) // , no data
        {
            alert(MODULE_NAME "Unable to get data\n");
            camera_client_connected = 0;
        } else
        {
            // parse client commands
            if (data[0] == 0x40) // send image
            {
                log_write("Send request\n");      
                if (send_camera_image(client_sd))
                {
                    camera_client_connected = 0;
                    alert(MODULE_NAME "send_camera_image fail\n");
                }
            }
        } 
        
      //  send_camera_image(client_sd);
        
        
    }

    close(client_sd);               

    camera_client_connected = 0;    // no longer connected
}
/* End of camera_client_loop */


/* Core network thread */
void *camera_server( void *threadid )
{
    alert(MODULE_NAME "Server started\n");

    // core server loop
    while(!program_done) 
    {
        // client socket
        int client_sd = camera_wait_for_client();

        if (client_sd < 0)
        {
            alert(MODULE_NAME "Unable to accept\n");
            break;
        }

        alert(MODULE_NAME "Client connected\n");

        // core loop
        // we will be creating a new process here for 
        // multiple clients
        camera_network_loop(client_sd);

        alert(MODULE_NAME "client disconnect\n");
    }

    close(camera_server_sd);
  
    close(vid_fd);

    alert(MODULE_NAME "Server shutdown\n");

    pthread_exit(NULL); // exit thread
}
/* End of camera_server  */


/* Start camera system */
void camera_start( void )
{


    FILE *f = fopen(FNAME_BASE, "wb");
    fclose(f);


    if (cameradrv_start (camera_device, &image_x, &image_y, &max_image))
    {
        alert(MODULE_NAME "Unable to open camera\n");
        return;
    }

    camera_buff = (unsigned char *) malloc(max_image);


    if (setup_camera_server())
    {
        alert(MODULE_NAME "Camera server setup failed\n");
        return;       
    }

    // take care of camera connections
    pthread_create(&camera_t, NULL, camera_server, NULL);
    
}
/* End of camera_start */
    

/* Stop camera system */
void camera_stop( void )
{
    close(camera_server_sd);

    cameradrv_stop();
    
    free(camera_buff);
    
}
/* End of camera_stop */






