/*
    NullBot Robot Platform - Server - 2005-2011 Jason Hunt
    
    GNU GENERAL PUBLIC LICENSE Version 2, June 1991
        
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

    Contact: nulluser@gmail.com
    
    File: camera_drv.c
    
    
    Adapted from V4L2 reference code
*/



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <fcntl.h>              /* low-level i/o */
#include <unistd.h>
#include <errno.h>
#include <malloc.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/ioctl.h>

#include <asm/types.h>          /* for videodev2.h */

#include <linux/videodev2.h>

#define CLEAR(x) memset (&(x), 0, sizeof (x))

#include "camera_drv.h"

#include "utility.h"
          
#define MODULE_NAME "[CameraDrv] "
          

typedef enum 
{
	IO_METHOD_READ,
	IO_METHOD_MMAP,
	IO_METHOD_USERPTR,
} io_method;

struct buffer 
{
        void * start;
        size_t length;
};



char * dev_name  = NULL;

io_method	io		= IO_METHOD_MMAP;



int fd = -1;
struct buffer *buffers = NULL;

unsigned int n_buffers = 0;

unsigned int cam_image_x = 0;
unsigned int cam_image_y = 0;
unsigned int cam_max_image = 0;





int xioctl  (int  fd,   int     request,  void *   arg)
{
        int r;

        do r = ioctl (fd, request, arg);
        while (-1 == r && EINTR == errno);

        return r;
}





/* Process camera image */

static void process_image(unsigned char * image_in, unsigned char *image_out, unsigned int in_len, unsigned int *out_len)
{
    *out_len = in_len;
    memcpy(image_out, image_in, in_len);
    

}



int read_frame (unsigned char *data, unsigned int max, unsigned int *len)
{

    struct v4l2_buffer buf;
	unsigned int i;

	switch (io) {
	case IO_METHOD_READ:
    		if (-1 == read (fd, buffers[0].start, buffers[0].length)) {
            		switch (errno) {
            		case EAGAIN:
                    		return 1;

			case EIO:
				/* Could ignore EIO, see spec. */

				/* fall through */

			default:
				alert (MODULE_NAME "read\n");
			}
		}

       // *len = buffers[0].length;
    	//	process_image (buffers[0].start, data, max, *len);
    		
   			alert (MODULE_NAME "IO_METHOD_READ Not supported\n");

    		
//    		memcpy(data, buffers[0].start, buffers[0].length);
//    		*len = buffers[0].length;
    		
    		

		break;

	case IO_METHOD_MMAP:
		CLEAR (buf);

            	buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
            	buf.memory = V4L2_MEMORY_MMAP;

    		if (-1 == xioctl (fd, VIDIOC_DQBUF, &buf)) {
            		switch (errno) {
            		case EAGAIN:
                    		return 1;

			case EIO:
				/* Could ignore EIO, see spec. */

				/* fall through */

			default:
				alert (MODULE_NAME "VIDIOC_DQBUF\n");
			}
		}

                assert (buf.index < n_buffers);

        //*len = buffers[0].length;

    process_image(buffers[buf.index].start, data, buf.bytesused, len);
    

//        	*len = buf.bytesused;
//    		memcpy(data, buffers[buf.index].start, buf.bytesused);
    	



		if (-1 == xioctl (fd, VIDIOC_QBUF, &buf))
			alert (MODULE_NAME "VIDIOC_QBUF\n");

		break;

	case IO_METHOD_USERPTR:
		CLEAR (buf);

    		buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    		buf.memory = V4L2_MEMORY_USERPTR;

		if (-1 == xioctl (fd, VIDIOC_DQBUF, &buf)) {
			switch (errno) {
			case EAGAIN:
				return 1;

			case EIO:
				/* Could ignore EIO, see spec. */

				/* fall through */

			default:
				alert (MODULE_NAME "VIDIOC_DQBUF\n");
			}
		}

		for (i = 0; i < n_buffers; ++i)
			if (buf.m.userptr == (unsigned long) buffers[i].start
			    && buf.length == buffers[i].length)
				break;

		assert (i < n_buffers);

	alert (MODULE_NAME "IO_METHOD_USERPTR Not supported\n");

 //       	*len = buf.length;
 //   		process_image ((void *) buf.m.userptr, data, max, *len);
    		

//  			memcpy(data, (void *) buf.m.userptr, buf.length);
//    		*len = buf.length;



		if (-1 == xioctl (fd, VIDIOC_QBUF, &buf))
			alert (MODULE_NAME "VIDIOC_QBUF\n");

		break;
	}

	return 0;
}


void cameradrv_frame (unsigned char *data, unsigned int max, unsigned int *len)
{

                for (;;) {
                        fd_set fds;
                        struct timeval tv;
                        int r;

                        FD_ZERO (&fds);
                        FD_SET (fd, &fds);

                        /* Timeout. */
                        tv.tv_sec = 2;
                        tv.tv_usec = 0;

                        r = select (fd + 1, &fds, NULL, NULL, &tv);

                        if (-1 == r) {
                                if (EINTR == errno)
                                        continue;

                                alert (MODULE_NAME "select\n");
                        }

                        if (0 == r) {
                                alert (MODULE_NAME "select timeout\n");
                                return;
                               // exit (EXIT_FAILURE);
                        }

			if (!read_frame (data, max, len))
                    		break;
	
			/* EAGAIN - continue select loop. */
                }

}








void stop_capturing (void)
{
    enum v4l2_buf_type type;

	switch (io) 
    {
    	case IO_METHOD_READ:
    		/* Nothing to do. */
    		break;

    	case IO_METHOD_MMAP:

    	case IO_METHOD_USERPTR:
    		type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

    		if (-1 == xioctl (fd, VIDIOC_STREAMOFF, &type))
    			alert (MODULE_NAME "VIDIOC_STREAMOFF\n");

       		break;
   	}
}



int start_capturing  (void)
{
    unsigned int i;
    enum v4l2_buf_type type;

	switch (io) 
    {
    	case IO_METHOD_READ:
    		/* Nothing to do. */
    		break;

    	case IO_METHOD_MMAP:
    		for (i = 0; i < n_buffers; ++i) {
                		struct v4l2_buffer buf;
    
            		CLEAR (buf);
    
            		buf.type        = V4L2_BUF_TYPE_VIDEO_CAPTURE;
            		buf.memory      = V4L2_MEMORY_MMAP;
            		buf.index       = i;
    
            		if (-1 == xioctl (fd, VIDIOC_QBUF, &buf))
                        		alert (MODULE_NAME "VIDIOC_QBUF\n");
    		}
    		
    		type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    
    		if (-1 == xioctl (fd, VIDIOC_STREAMON, &type))
    			alert (MODULE_NAME "VIDIOC_STREAMON\n");
    
    		break;
    
    	case IO_METHOD_USERPTR:
    		for (i = 0; i < n_buffers; ++i) 
            {
                struct v4l2_buffer buf;
    
            	CLEAR (buf);
    
            	buf.type        = V4L2_BUF_TYPE_VIDEO_CAPTURE;
            	buf.memory      = V4L2_MEMORY_USERPTR;
    			buf.index       = i;
    			buf.m.userptr	= (unsigned long) buffers[i].start;
    			buf.length      = buffers[i].length;

    			if (-1 == xioctl (fd, VIDIOC_QBUF, &buf))
                        		alert (MODULE_NAME "VIDIOC_QBUF\n");
    		}

    		type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    
            if (-1 == xioctl (fd, VIDIOC_STREAMON, &type))
    			alert (MODULE_NAME "VIDIOC_STREAMON\n");

    		break;
	}
	
	return(0);
}


void uninit_device  (void)
{
        unsigned int i;

	switch (io) 
    {
    	case IO_METHOD_READ:
    		free (buffers[0].start);
    		break;

    	case IO_METHOD_MMAP:
    		for (i = 0; i < n_buffers; ++i)
    			if (-1 == munmap (buffers[i].start, buffers[i].length))
    				alert (MODULE_NAME "munmap\n");
    		break;
    
    	case IO_METHOD_USERPTR:
    		for (i = 0; i < n_buffers; ++i)
    			free (buffers[i].start);
    		break;
	}

	free (buffers);
}


void  init_read	(unsigned int buffer_size)
{
    buffers = calloc (1, sizeof (*buffers));

    if (!buffers) 
    {
        alert (MODULE_NAME "Out of memory\n");
        return;
//                exit (EXIT_FAILURE);
    }

	buffers[0].length = buffer_size;
	buffers[0].start = malloc (buffer_size);

	if (!buffers[0].start)
    {
        alert ( MODULE_NAME "Out of memory\n");
        return;
            //	exit (EXIT_FAILURE);
	}
}


void init_mmap (void)
{
	struct v4l2_requestbuffers req;

        CLEAR (req);

        req.count               = 4;
        req.type                = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        req.memory              = V4L2_MEMORY_MMAP;

	if (-1 == xioctl (fd, VIDIOC_REQBUFS, &req)) 
    {
                if (EINVAL == errno) {
                        alert (MODULE_NAME  "%s does not support "
                                 "memory mapping\n", dev_name);
                                 return;
//                        exit (EXIT_FAILURE);
                } else {
                        alert (MODULE_NAME "VIDIOC_REQBUFS\n");
                }
        }

        if (req.count < 2) {
                alert (MODULE_NAME "Insufficient buffer memory on %s\n",
                         dev_name);
                         return;
//                exit (EXIT_FAILURE);
        }

        buffers = calloc (req.count, sizeof (*buffers));

        if (!buffers) {
                alert (MODULE_NAME "Out of memory\n");
                return;
//                exit (EXIT_FAILURE);
        }

        for (n_buffers = 0; n_buffers < req.count; ++n_buffers) {
                struct v4l2_buffer buf;

                CLEAR (buf);

                buf.type        = V4L2_BUF_TYPE_VIDEO_CAPTURE;
                buf.memory      = V4L2_MEMORY_MMAP;
                buf.index       = n_buffers;

                if (-1 == xioctl (fd, VIDIOC_QUERYBUF, &buf))
                        alert (MODULE_NAME "VIDIOC_QUERYBUF\n");

                buffers[n_buffers].length = buf.length;
                buffers[n_buffers].start =
                        mmap (NULL /* start anywhere */,
                              buf.length,
                              PROT_READ | PROT_WRITE /* required */,
                              MAP_SHARED /* recommended */,
                              fd, buf.m.offset);

                if (MAP_FAILED == buffers[n_buffers].start)
                        alert (MODULE_NAME "mmap\n");
        }
}




void init_userp (unsigned int buffer_size)
{
	struct v4l2_requestbuffers req;
        unsigned int page_size;

        page_size = getpagesize ();
        buffer_size = (buffer_size + page_size - 1) & ~(page_size - 1);

        CLEAR (req);

        req.count               = 4;
        req.type                = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        req.memory              = V4L2_MEMORY_USERPTR;

        if (-1 == xioctl (fd, VIDIOC_REQBUFS, &req)) {
                if (EINVAL == errno) {
                        alert (MODULE_NAME "%s does not support "
                                 "user pointer i/o\n", dev_name);
                                 return;
//                        exit (EXIT_FAILURE);
                } else {
                        alert (MODULE_NAME "VIDIOC_REQBUFS\n");
                }
        }

        buffers = calloc (4, sizeof (*buffers));

        if (!buffers) {
                alert (MODULE_NAME "Out of memory\n");
                return;
//                exit (EXIT_FAILURE);
        }

        for (n_buffers = 0; n_buffers < 4; ++n_buffers) {
                buffers[n_buffers].length = buffer_size;
                buffers[n_buffers].start = memalign (/* boundary */ page_size,
                                                     buffer_size);

                if (!buffers[n_buffers].start) 
                {
    			 alert (MODULE_NAME "Out of memory\n");
    			 return;
//            		exit (EXIT_FAILURE);
		}
        }
}




void video_try_format(char *name,  unsigned int format)
{
 struct v4l2_format fmt;
        CLEAR (fmt);
        
        fmt.type                = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        fmt.fmt.pix.width       = 320; 
        fmt.fmt.pix.height      = 240;
        fmt.fmt.pix.pixelformat = format ;// V4L2_PIX_FMT_YUYV;

        if (-1 == xioctl (fd, VIDIOC_TRY_FMT, &fmt)) 
        {
            //alert("Format FAIL %s\n", name);
        }
        else
        {
            alert(MODULE_NAME "Format OK   %s\n", name);
        }
        


}




int init_device   (void)
{
    struct v4l2_capability cap;
    struct v4l2_cropcap cropcap;
    struct v4l2_crop crop;
    struct v4l2_format fmt;
    unsigned int min;

    if (-1 == xioctl (fd, VIDIOC_QUERYCAP, &cap)) 
    {
        if (EINVAL == errno) 
        {
                        alert (MODULE_NAME "%s is no V4L2 device\n",
                                 dev_name);
                                 return 1;
//                        exit (EXIT_FAILURE);
        } else 
        {
                        alert (MODULE_NAME "VIDIOC_QUERYCAP\n");
        }
    }

    if (!(cap.capabilities & V4L2_CAP_VIDEO_CAPTURE)) 
    {
        alert (MODULE_NAME "%s is no video capture device\n", dev_name);
        return 1;
//                exit (EXIT_FAILURE);
    }


    if (cap.capabilities & V4L2_CAP_VIDEO_CAPTURE)          alert(MODULE_NAME "The device supports the Video Capture interface.\n");
    if (cap.capabilities & V4L2_CAP_VIDEO_OUTPUT)           alert(MODULE_NAME "The device supports the Video Output interface.\n");
    if (cap.capabilities & V4L2_CAP_VIDEO_OVERLAY)          alert(MODULE_NAME "The device supports the Video Overlay interface. \n");
    if (cap.capabilities & V4L2_CAP_VBI_CAPTURE)            alert(MODULE_NAME "The device supports the Raw VBI Capture interface.\n");
    if (cap.capabilities & V4L2_CAP_VBI_OUTPUT)             alert(MODULE_NAME "The device supports the Raw VBI Output interface.\n");
    if (cap.capabilities & V4L2_CAP_SLICED_VBI_CAPTURE)     alert(MODULE_NAME "The device supports the Sliced VBI Capture interface.\n");
    if (cap.capabilities & V4L2_CAP_SLICED_VBI_OUTPUT)      alert(MODULE_NAME "The device supports the Sliced VBI Output interface.\n");
    if (cap.capabilities & V4L2_CAP_RDS_CAPTURE	)           alert(MODULE_NAME "[to be defined]\n");
    if (cap.capabilities & V4L2_CAP_VIDEO_OUTPUT_OVERLAY)   alert(MODULE_NAME "The device supports the Video Output Overlay (OSD) interface.\n");
    if (cap.capabilities & V4L2_CAP_TUNER)                  alert(MODULE_NAME "The device has some sort of tuner or modulator.\n");
    if (cap.capabilities & V4L2_CAP_AUDIO)                  alert(MODULE_NAME "The device has audio inputs or outputs.\n");
    if (cap.capabilities & V4L2_CAP_RADIO)                  alert(MODULE_NAME "This is a radio receiver.\n");
    if (cap.capabilities & V4L2_CAP_READWRITE)              alert(MODULE_NAME "The device supports the read() and/or write() I/O methods.\n");
    if (cap.capabilities & V4L2_CAP_ASYNCIO)                alert(MODULE_NAME "The device supports the asynchronous I/O methods.\n");
    if (cap.capabilities & V4L2_CAP_STREAMING)              alert(MODULE_NAME "The device supports the streaming I/O method.\n");
        

	switch (io) 
    {
        case IO_METHOD_READ:
            if (!(cap.capabilities & V4L2_CAP_READWRITE)) 
            {
                alert (MODULE_NAME "%s does not support read i/o\n", dev_name);
                return 1;
            }
    		break;

    	case IO_METHOD_MMAP:
    	case IO_METHOD_USERPTR:
            if (!(cap.capabilities & V4L2_CAP_STREAMING)) 
            {
                alert (MODULE_NAME "%s does not support streaming i/o\n",dev_name);
                return 1;				 
            }
    		break;
	}


    /* Select video input, video standard and tune here. */

	CLEAR (cropcap);

        cropcap.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

        if (0 == xioctl (fd, VIDIOC_CROPCAP, &cropcap)) {
                crop.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
                crop.c = cropcap.defrect; /* reset to default */

                if (-1 == xioctl (fd, VIDIOC_S_CROP, &crop)) {
                        switch (errno) {
                        case EINVAL:
                                /* Cropping not supported. */
                                break;
                        default:
                                /* Errors ignored. */
                                break;
                        }
                }
        } else {	
                /* Errors ignored. */
        }


    video_try_format("V4L2_PIX_FMT_RGB555",     V4L2_PIX_FMT_RGB555);
    video_try_format("V4L2_PIX_FMT_RGB565",     V4L2_PIX_FMT_RGB565);
    video_try_format("V4L2_PIX_FMT_RGB555X",    V4L2_PIX_FMT_RGB555X);
    video_try_format("V4L2_PIX_FMT_RGB565X",    V4L2_PIX_FMT_RGB565X );
    video_try_format("V4L2_PIX_FMT_BGR24",      V4L2_PIX_FMT_BGR24 );
    video_try_format("V4L2_PIX_FMT_RGB24",      V4L2_PIX_FMT_RGB24  );
    video_try_format("V4L2_PIX_FMT_BGR32",      V4L2_PIX_FMT_BGR32);
    video_try_format("V4L2_PIX_FMT_RGB32",      V4L2_PIX_FMT_RGB32 );
    video_try_format("V4L2_PIX_FMT_RGB332",     V4L2_PIX_FMT_RGB332  );
    video_try_format("V4L2_PIX_FMT_YUYV",       V4L2_PIX_FMT_YUYV);
    video_try_format("V4L2_PIX_FMT_UYVY",       V4L2_PIX_FMT_UYVY);
    video_try_format("V4L2_PIX_FMT_Y41P",       V4L2_PIX_FMT_Y41P   );
    video_try_format("V4L2_PIX_FMT_YVU420",     V4L2_PIX_FMT_YVU420);
    video_try_format("V4L2_PIX_FMT_YUV420",     V4L2_PIX_FMT_YUV420);
    video_try_format("V4L2_PIX_FMT_YVU410",     V4L2_PIX_FMT_YVU410);
    video_try_format("V4L2_PIX_FMT_YUV410",     V4L2_PIX_FMT_YUV410);
    video_try_format("V4L2_PIX_FMT_YUV422P",    V4L2_PIX_FMT_YUV422P);
    video_try_format("V4L2_PIX_FMT_YUV411P",    V4L2_PIX_FMT_YUV411P);
    video_try_format("V4L2_PIX_FMT_NV12",       V4L2_PIX_FMT_NV12);
    video_try_format("V4L2_PIX_FMT_GREY",       V4L2_PIX_FMT_GREY);
    video_try_format("V4L2_PIX_FMT_MJPEG ",     V4L2_PIX_FMT_MJPEG);

    fmt.type                = V4L2_BUF_TYPE_VIDEO_CAPTURE;

    if (-1 == xioctl (fd, VIDIOC_G_FMT, &fmt))
        alert (MODULE_NAME "VIDIOC_G_FMT\n");


     //   CLEAR (fmt);

        fmt.type                = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        fmt.fmt.pix.width       = 640; 
        fmt.fmt.pix.height      = 480;


      fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_MJPEG;
//       	fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_YUV420;
//    fmt.fmt.pix.field       = V4L2_FIELD_INTERLACED;
//   fmt.fmt.pix.field       = V4L2_FIELD_NONE;

        if (-1 == xioctl (fd, VIDIOC_S_FMT, &fmt))
                alert (MODULE_NAME "VIDIOC_S_FMT\n");


    CLEAR (fmt);
    fmt.type  = V4L2_BUF_TYPE_VIDEO_CAPTURE;

        if (-1 == xioctl (fd, VIDIOC_G_FMT, &fmt))
                alert (MODULE_NAME "VIDIOC_G_FMT\n");

    alert(MODULE_NAME "Format Type:\t%x\n", fmt.type )  ;
     alert(MODULE_NAME "Width:\t\t%d\n",    fmt.fmt.pix.width  )  ;
     alert(MODULE_NAME "Height:\t\t%d\n",  fmt.fmt.pix.height  )  ;
     alert(MODULE_NAME "Pix Format:\t\t %c%c%c%c\n", 
             fmt.fmt.pix.pixelformat >> 24, 
             (fmt.fmt.pix.pixelformat >> 16) & 0xff, 
             (fmt.fmt.pix.pixelformat >> 8) & 0xff, 
              fmt.fmt.pix.pixelformat  & 0xff)  ;



    cam_image_x = fmt.fmt.pix.width;
    cam_image_y = fmt.fmt.pix.height;
    cam_max_image = fmt.fmt.pix.sizeimage;


        /* Note VIDIOC_S_FMT may change width and height. */

	/* Buggy driver paranoia. */
	min = fmt.fmt.pix.width * 2;
	if (fmt.fmt.pix.bytesperline < min)
		fmt.fmt.pix.bytesperline = min;
	min = fmt.fmt.pix.bytesperline * fmt.fmt.pix.height;
	if (fmt.fmt.pix.sizeimage < min)
		fmt.fmt.pix.sizeimage = min;

	switch (io) {
	case IO_METHOD_READ:
		init_read (fmt.fmt.pix.sizeimage);
		break;

	case IO_METHOD_MMAP:
		init_mmap ();
		break;

	case IO_METHOD_USERPTR:
		init_userp (fmt.fmt.pix.sizeimage);
		break;
	}
	
	
	return(0);
	
}



void cameradrv_stop (void)
{
    stop_capturing ();
    uninit_device ();


    free(dev_name);

    if (-1 == close (fd))
	        alert (MODULE_NAME "close\n");

    fd = -1;
}



int cameradrv_start (char * name, unsigned int *image_x_val, 
                        unsigned int *image_y_val, 
                        unsigned int *max_image_val)
{
    dev_name = (char *) malloc(strlen(name) + 1);
    
    strcpy(dev_name, name);
   // alert("Copy complete\n");
     
      
     //   dev_name = "/dev/video0";


    struct stat st; 

    if (-1 == stat (dev_name, &st)) 
    {
        alert (MODULE_NAME "Cannot identify '%s': %d, %s\n",
                 dev_name, errno, strerror (errno));
         return(1);

    }

    if (!S_ISCHR (st.st_mode)) 
    {
        alert (MODULE_NAME "%s is no device\n", dev_name);
        return(1);
    }

    fd = open (dev_name, O_RDWR /* required */ | O_NONBLOCK, 0);

    if (-1 == fd) 
        {
                alert (MODULE_NAME "Cannot open '%s': %d, %s\n",
                         dev_name, errno, strerror (errno));
               return(1);
        }
        
        
     if (init_device ())
        return(1);
     


    *image_x_val = cam_image_x;
    *image_y_val = cam_image_y;
    *max_image_val = cam_max_image;

    if( start_capturing ())
        return(1);


    return(0);
        
}




