/*
    Robot Controller Client

    (C) 2005-2011 Jason Hunt

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

    Contact: nulluser@gmail.com

    FILE: automode.cpp
        
    Automatic control of a robot
*/

#include <windows.h>
#include <stdio.h>
#include <math.h>

#include "types.h"
#include "robot_win.h"
#include "robot.h"
#include "main.h"
#include "world.h"
#include "utility.h"
#include "automode.h"

#define MODULE_NAME "[Automode] "


#define ir_fl  ir[0]
#define ir_f   ir[1]
#define ir_fr  ir[2]


/* Compute nav forward */
void robot_class::compute_forward ( void )
{
    actions[0].dir = 0;
    actions[0].speed = 0.1;
    actions[0].pri =  0.1;
}
/* End of compute_forward */


    
/* Compute nav reverse */    
void robot_class::compute_reverse ( void )
{
 //  double ir_fl = robot.ir[0];
  //  double ir_f =  robot.ir[1];
  //  double ir_fr = robot.ir[2];


    
    actions[2].dir =  turn_bias;
    actions[2].speed = -0.05;////0.1;
    actions[2].pri =  rev_bias;


    char b[300];
    sprintf(b, "ir: %f %f %f\n", ir_fl,ir_f,ir_fr);
   // write_console(b);

        // reverse needed
    if (ir_f > 1.6)
    {
        turn_bias = ir_fl - ir_fr + f_rand(-1,1);
        rev_bias = 1;
    }
    
    rev_bias /= 1.2;
    turn_bias /= 1.1;
}
/* End of compute reverse */


/* Compute nav avoid */
void robot_class::compute_avoid ( void )
{
    //double ir_fl = robot.ir[0];
   // double ir_f =  robot.ir[1];
   // double ir_fr = robot.ir[2];

    
    actions[1].dir = ir_fl - ir_fr;
    actions[1].speed = 0;//0.1;
    actions[1].pri = (ir_fl + ir_fr) ;    
}
/* End of comoute avoid */





/* Setup or check a timer */
double robot_class::timer_func(double v, unsigned int timer_num, double threshhold, double lim)
{
    if (timer_num >= NUM_TIMERS) return(0);

    double timer_time = 1.0;


    
    if (first_run)
    {
        for (int i = 0; i < NUM_TIMERS; i++)
        {
            stime[i] = 0;
            limits[NUM_TIMERS];
        }

        first_run = 0;
    }
    

    // Timer already running
    if (stime[timer_num] > 0)
    {
         double time_diff = get_time() - stime[timer_num];

        // Timer done?
        if (time_diff > timer_time)
        {            
            stime[timer_num] = 0;
            return(1);
        }

        return(limits[timer_num] * time_diff / timer_time);

    } else
    if (v >= threshhold)   // Start timer
    {
        stime[timer_num] = get_time();
        limits[timer_num] = lim;
        return(0);            
    }         
    return(0);
}
/* End of timer_func */


/* Compute neural network for automode */
void robot_class::robot_auto_mode( void )
{
    
    for (int i = 0; i < AVERAGE_SIZE - 1; i++)
    {
        nav_average[i] = nav_average[i+1];
    }
    nav_average[AVERAGE_SIZE - 1] = nav_off;
    
    
    double nav_ave = 0;
    for (int i = 0; i < AVERAGE_SIZE; i++)
        nav_ave += nav_average[i];

    nav_ave /=  (double)AVERAGE_SIZE;
    
    
    char b[300];
    sprintf(b, "ave; %f\n", nav_ave);
//    console(b);
    
 //   double ir_fl = robot.ir[0];
   // double ir_f =  robot.ir[1];
   // double ir_fr = robot.ir[2];
            
    
    double sl = touch[0];
    double sf = touch[1];
    double sr = touch[2];

    double rev_sum = 0.2*ir_fl + 0.2*ir_fr + 0.4*ir_f + 0.4*sl + 0.4 * sr + 2.0 * sf;

  
    double rev_time_mod = timer_func(rev_sum, 0, 2, 2);
    double turn_time_mod = timer_func(rev_time_mod , 1, 1, f_rand(-2,2));


       double mag = nav_mag;  
       if (mag < 0.5) mag = 0.5;
       
    double ms =  1 - 0.15*ir_fl - 0.15*ir_fl - 1.1*rev_time_mod;
//    double ms = 1 - sl - sr - 4*rev_time_mod;

    // 2.5 turn mag works

    double mr = 1.5* nav_ave + 0.50*ir_fl - 0.50*ir_fr - 1.0*turn_time_mod;

//    double mr = 1.5* robot.nav_off + 0.50*ir_fl - 0.50*ir_fr - 1.0*turn_time_mod;

//    double mr = sl - sr - 2*turn_time_mod;    


    double n_ddir = 2.0 * mr;
    double n_speed = 1.0 *  ms;

if (n_speed < 0) n_speed = 0;

    motion(n_speed, n_ddir);

}
/* End of robot_auto_mode */














    
