/*
    Robot Controller Client

    (C) 2005-2011 Jason Hunt

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

    Contact: nulluser@gmail.com

    FILE: graphics.h
    
    Graphics primitives    
*/

#include <windows.h>
#include <stdio.h>
#include <math.h>


#include "types.h"
#include "robot.h"
#include "world.h"


#define MODULE_NAME "[Graphics] "




/* Draw arrow */
/* GDI code to draw a vector graphics arrow */
void world_class::draw_arrow(HDC dc, double x, double y, double dir, double size)
{
    double arrow_base_x = x - size * cos(dir);
    double arrow_base_y = y + size * sin(dir);
    
    MoveToEx(dc, get_screen_x(arrow_base_x), get_screen_y(arrow_base_y), NULL);
    
    double arrow_tip_x = x + size * cos(dir);
    double arrow_tip_y = y - size * sin(dir);
        
    LineTo(dc, get_screen_x(arrow_tip_x), get_screen_y(arrow_tip_y));    

    double arrow_left_x = x + 0.75 * size * cos(dir - M_PI / 4.0);
    double arrow_left_y = y - 0.75 * size * sin(dir - M_PI / 4.0);

    MoveToEx(dc, get_screen_x(arrow_left_x), get_screen_y(arrow_left_y), NULL);

    LineTo(dc, get_screen_x(arrow_tip_x), get_screen_y(arrow_tip_y));    

    double arrow_right_x = x + 0.75 * size * cos(dir + M_PI / 4.0);
    double arrow_right_y = y - 0.75 * size * sin(dir + M_PI / 4.0);

    LineTo(dc, get_screen_x(arrow_right_x), get_screen_y(arrow_right_y));    
}
/* End of draw arrow */


/* Draw line in world cords */
void world_class::draw_line(HDC dc, line l)
{   
    MoveToEx(dc, get_screen_x(l.x1), get_screen_y(l.y1), NULL);
    LineTo(dc, get_screen_x(l.x2), get_screen_y(l.y2));    
}
/* End of draw line */


/* Draw a sensor for an animal */
void world_class::draw_sensor(HDC dc, double x, double y, double dir, double dist, double size, COLORREF c)
{
    double sx = x + dist * cos(dir);
    double sy = y - dist * sin(dir);
    
    RECT r;
    
    r.left = get_screen_x(sx - size / 2.0);
    r.top = get_screen_y(sy - size / 2.0);
    r.right = get_screen_x(sx + size / 2.0);
    r.bottom = get_screen_y(sy + size / 2.0);
    
    HBRUSH b = CreateSolidBrush(c);

    HBRUSH hold = (HBRUSH) SelectObject(dc, b);

    Ellipse(dc, r.left, r.top, r.right, r.bottom);    
    
    SelectObject(dc, hold);    
        
    DeleteObject(b);
}
/* End of draw sensor */


/* Decode the sensor into a color */
COLORREF get_sensor_color( bool val )
{
    if (val) return(RGB(255, 0, 0));     // Red for well

    return(RGB(255, 255, 255));                 // No hit, return white
}
/* End of get_sensor_color */













