/*
    Robot Controller Client

    (C) 2005-2011 Jason Hunt

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

    Contact: nulluser@gmail.com

    FILE: main.cpp
    
    Main window
*/

#include <windows.h>
#include <stdio.h>


#include "types.h"
#include "resource.h"
#include "main.h"
#include "graphics.h"
#include "robot.h"
#include "world.h"
#include "..\version.h"

#define MODULE_NAME "[Main] "

/* Main class constructor */
main_class::main_class( void )
{
    log (MODULE_NAME "constuctor\n");

    memset(keys, 0, sizeof(keys));              // Clear key array

    updates = 0;                // Number of updates since last second
    updates_per_second = 0;     // Number of updates per second

    manual_mode = true;

    timer_val = 10;

    mouse_is_down = false;
    right_mouse_down = false;
    show_stats = false;
   
    mouse_is_down = false;
    right_mouse_down  = false;
    
   
    // Create brushs
    white_brush = CreateSolidBrush(RGB(255, 255, 255));
    black_brush = CreateSolidBrush(RGB(0, 0, 0));
    green_brush = CreateSolidBrush(RGB(0, 255, 0));
    blue_brush = CreateSolidBrush(RGB(0, 0, 255));
    red_brush = CreateSolidBrush(RGB(255, 0, 0));
    color1_brush = CreateSolidBrush(RGB(0, 128, 200));
    

    // Create pens
    white_pen = CreatePen(PS_SOLID, 1, RGB(255, 255, 255));
    black_pen = CreatePen(PS_SOLID, 1, RGB(0, 0, 0));
    green_pen = CreatePen(PS_SOLID, 1, RGB(0, 255, 0));
    blue_pen = CreatePen(PS_SOLID, 1, RGB(0, 0, 255));
    color1_pen = CreatePen(PS_SOLID, 1, RGB(0, 128, 200));    

}
/* End of main class constructor */


/* Main class destructor */
main_class::~main_class( void )
{
    DeleteObject(black_brush);
    DeleteObject(white_brush);
    DeleteObject(green_brush);
    DeleteObject(blue_brush);
    DeleteObject(red_brush);
    DeleteObject(color1_brush);
    
    DeleteObject(white_pen);
    DeleteObject(black_pen);
    DeleteObject(green_pen);
    DeleteObject(blue_pen);
    DeleteObject(color1_pen);
}
/* End of main class destructor */


/* Init the entire system */
void main_class::main_startup( HWND hwnd )
{
    WSADATA wsda;
    WSAStartup(MAKEWORD(1, 1),&wsda);
   
    rand_seed();
     
    memdc_setup(hwnd);          // Setup buffer

    g_world = new world_class(this);
    
    g_world->update();
    g_world->draw(memdc);    


    SetTimer(hwnd, UPDATE_ID, timer_val, NULL);        
    SetTimer(hwnd, SECOND_ID, 1000, NULL);        
}
/* End of system_startup */


/* De-Init the entire system */
void main_class::main_shutdown( void )
{
    delete g_world;
            
    DeleteObject(mem_bitmap);
    DeleteDC(memdc);   
    
    WSACleanup();
}
/* End of system_shutdown */


/* Called when a key is pressed */
void main_class::key_down(HWND hwnd, WPARAM w)
{
    // Mark key as pressed
    keys[w & 0xff] = true;  

    // Deal with single press events here
    if (keys['N'])
    {
        g_world->free_markers();
        g_world->free_impacts();
    }
    
//    if (keys['V']) world.robot.toggle_vision();

    if (keys[27]) PostQuitMessage(0);


    if (keys[VK_DOWN])
    {
        timer_val++;
        
        SetTimer(hwnd, UPDATE_ID, timer_val, NULL);                     
    }
    
    if (keys[VK_UP])
    {
        timer_val --;
        
        if (timer_val < 0) timer_val = 0;
        
       SetTimer(hwnd, UPDATE_ID, timer_val, NULL);
    }
    
    

    //if (keys['R']) world.setup(); 
                  
    if (keys['T']) g_world-> toggle_walls();

    if (keys['M']) g_world->toggle_show_markers(); 
    if (keys['I']) g_world->toggle_show_impacts(); 
    if (keys['H']) g_world->toggle_show_map();
    

     if (keys['G']) g_world->toggle_show_grad();
     
     
    /* if (keys['1']) g_world->scale1 -= 0.01;
     if (keys['2']) g_world->scale1 += 0.01;

     if (keys['3']) g_world->scale2 -= 0.01;
     if (keys['4']) g_world->scale2 += 0.01;
     
      if (keys['5']) g_world->scale3 -= 0.001;
     if (keys['6']) g_world->scale3 += 0.001;*/
     
     if (keys[VK_SPACE]) show_stats = !show_stats;
     
    // if (keys['P']) g_world->world_compute_map_grid();
}
/* End of key_down */


/* Deal with main menu commands */
void main_class::main_menu(HWND hwnd, unsigned int id)
{


    switch(id)
    {
      /*  case ID_EXIT:
            DestroyWindow(hwnd);
            return;

        case ID_SAVE:
            world_save();
            return;

        case ID_OPEN:
            world_load();
            return;*/
            
    }    
    
}
/* End of main menu */


/* Called when a key is released */
void main_class::key_up(WPARAM w)
{
    keys[w & 0xff] = false;            
}
/* End of key up */


/* Process currently pressed keys */
void main_class::check_keys( void )
{
    
        
}
/* End of check keys */


/* Process mouse down */
void main_class::mouse_down(LPARAM l)
{
    mouse_is_down = true; 
    
//    double world_x1 = (LOWORD(l) / (double)map_display_x) * g_world->x_size;
//    double world_y1 = (HIWORD(l) / (double)map_display_y) * g_world->y_size;

    double world_x1 = g_world->get_world_x( LOWORD(l));    // Convert Screen to world pos
    double world_y1 = g_world->get_world_y( HIWORD(l));

    g_world->add_wall(world_x1, world_y1, 0, 0);
}
/* End of mouse down */


/* Process mouse up */
void main_class::mouse_up(LPARAM l)
{
    mouse_is_down = false;
}
/* Process mouse up */


/* Process mouse move */
void main_class::mouse_move(LPARAM l)
{
    if (mouse_is_down)
    {    
        if (!g_world->wall_present()) return;

        double world_x2 = g_world->get_world_x( LOWORD(l));    // Convert Screen to world pos
        double world_y2 = g_world->get_world_y( HIWORD(l));

        if (world_x2 < g_world->get_wall0_x()) return;
        if (world_y2 < g_world->get_wall0_y()) return;
    
        g_world->set_wall0_size(world_x2 - g_world->get_wall0_x(), world_y2 - g_world->get_wall0_y());
    
    
    }    
    
    if (right_mouse_down)
    {
            g_world->set_target(g_world->get_world_x(LOWORD(l)),
                             g_world->get_world_y(HIWORD(l))); 
    
    }
}
/* End of mouse move */


/* Deal with the system timer */
void main_class::timer_message(HWND hwnd, WPARAM wparam)
{
    if (wparam == UPDATE_ID)
    {
        check_keys();

        g_world->update();
        g_world->draw(memdc);
        
        updates++;
        InvalidateRect(hwnd, NULL, 0);                            
        return;
    }

    if (wparam == SECOND_ID)
    {
        updates_per_second = updates;
        updates = 0;    
    }            
}
/* End of timer_message */


/* Deal with paint message */
void main_class::window_paint(HWND hwnd)
{
    PAINTSTRUCT ps;
    HDC hdc = BeginPaint(hwnd, &ps);

    char buff[400];

    if (show_stats)
    {
        //sprintf(buff, "Update rate: %3d scale1 %f scale2 %f scale3: %f", 
//                updates_per_second, g_world->scale1, g_world->scale2, g_world->scale3);
                       
  //      TextOut(memdc, 1, 1, buff, strlen(buff)); 
    }

    // Copy memory device context to display 
    BitBlt(hdc, 0, 0, map_display_x, map_display_y, memdc, 0, 0, SRCCOPY);
            
    EndPaint(hwnd, &ps); 
    
  // InvalidateRect(neuron_window, NULL, 0);      
}
/* End of window_paint */


/* Setup the off screen drawing system */
void main_class::memdc_setup( HWND hwnd )
{
    RECT r;
    GetClientRect(hwnd, &r);

    // Update stored client area size
    map_display_x = r.right - r.left;
    map_display_y = r.bottom - r.top;

    log(MODULE_NAME "memdc_setup x:   %d  y: %d\n", map_display_x, map_display_y);


    HDC hdc = GetDC(hwnd);

    // Create memory device context and bitmap            
    memdc = CreateCompatibleDC(hdc);
    mem_bitmap = CreateCompatibleBitmap(hdc, r.right, r.bottom);
            
    // Select the bitmap into the memort dc
    SelectObject(memdc, mem_bitmap);
            
    ReleaseDC(hwnd, hdc);     
}
/* End of memdc_setup */


/* Wrapper for the camera window procedure */
LRESULT CALLBACK main_class::proc_wrapper(HWND hwnd, UINT message, WPARAM w, LPARAM l)
{
    // Get object pointer from window data and call actual wndproc
        
    // Set the class pointer on creation
    if (message == WM_NCCREATE)
    {     
        if(l)
        {
            LONG handle = (LONG)((CREATESTRUCT*)l)->lpCreateParams;
        
            if(handle)
                SetWindowLong(hwnd, GWL_USERDATA, handle);
        }
    }
    
    main_class *win = (main_class*)::GetWindowLong(hwnd, GWL_USERDATA);
            
    // Call the member function     
    win -> proc(hwnd, message, w, l);
}
/* End of cam_proc_wrapper */


/* Window function */
LRESULT CALLBACK main_class::proc(HWND hwnd, UINT message, WPARAM wparam, LPARAM lparam)
{
    switch(message)
    {
        case WM_TIMER: timer_message(hwnd, wparam); return(0);

        case WM_CREATE: main_startup(hwnd); return(0);

        case WM_PAINT: window_paint(hwnd); return(0);

        case WM_DESTROY: 
            main_shutdown();            
            PostQuitMessage(0);
            return(0);
                                        
        case WM_SIZE:               // Process size message
            // These are alreadt setup, delete them
            DeleteObject(mem_bitmap);
            DeleteDC(memdc);
            
            // Setup again with the new size
            memdc_setup(hwnd);
            return(0);
                
        case WM_KEYDOWN:            // Process keydown
            key_down(hwnd, wparam);
            return(0);
    
        case WM_KEYUP:              // Process keyup
            key_up(wparam);
            return(0);
                
        case WM_LBUTTONDOWN:
            mouse_down(lparam);
            return(0);
               
        case WM_LBUTTONUP:
            mouse_up(lparam);
            return(0);
                          
        case WM_MOUSEMOVE:
            mouse_move(lparam);
            return(0);
             
            
        case WM_RBUTTONDOWN:
            g_world->set_target(g_world->get_world_x(LOWORD(lparam)),
                             g_world->get_world_y(HIWORD(lparam))); 

            right_mouse_down = true;
            return(0);
                                      
        case WM_RBUTTONUP:
            right_mouse_down = false;
            return(0);
                                
        case WM_COMMAND:
            main_menu(hwnd, LOWORD(wparam));
            return(0);
                
        default:
            return(DefWindowProc(hwnd, message, wparam, lparam));    
    }
    return(0);
}
/* Wnd of window function */



/* Create the main window */
void main_class::setup_window( void )
{
    WNDCLASS wc;
    
    HINSTANCE  hinst = 0;
    int mode = 1;
    
    wc.hInstance = hinst;
    wc.lpszClassName = "SwarmWin";
    wc.lpfnWndProc = proc_wrapper;
    wc.style = 0;
    
    wc.hIcon = LoadIcon(NULL, IDI_APPLICATION);
    wc.hCursor = LoadCursor(NULL, IDC_ARROW);
    wc.lpszMenuName = "MAINMENU";
    wc.cbClsExtra = 0;
    wc.cbWndExtra = 0;
        
    wc.hbrBackground = NULL;
    
    char name[256];
    sprintf(name, "NullBot Swarm Controller. Version %s %s  ", VERSION, COPYRIGHT);
    
    if (!RegisterClass(&wc)) return ;
    


    HWND hwnd = CreateWindowEx (0, "SwarmWin",
                                name,  
                                WS_OVERLAPPEDWINDOW,
                                120, 120, // x, y
                                640,480, // height, width
                                HWND_DESKTOP, NULL,
                                NULL, this);
                            
    ShowWindow( hwnd, mode);
    UpdateWindow( hwnd );


}
/* End of setup_window */


/* Winmain */
int WINAPI WinMain(HINSTANCE hinstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
    main_class *g_main = new main_class;        // Create main 
    
    g_main->setup_window();                     // Create the main window
     
   
    // Message loop
    MSG messages;

    while (GetMessage (&messages, NULL, 0, 0))
    {
        TranslateMessage(&messages);
        DispatchMessage(&messages);
    }
    
    delete g_main;

    return(messages.wParam);
}
/* End of Winmain */






