/*
    Robot Controller Client

    (C) 2005-2011 Jason Hunt

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

    Contact: nulluser@gmail.com

    FILE: main.h
*/


#ifndef MAIN_H
#define MAIN_H

class main_class;
class robot_class;
class robot_comm_class;
class robot_win_class;
class robot_camera_class;


#include "robot.h"
#include "world.h"
#include "profile.h"



#define UPDATE_ID 101
#define SECOND_ID 102








class main_class 
{
    public:
    
    main_class( void );
    ~main_class( void );
    
    void setup_window( void );
    
    void main_startup( HWND hwnd );
    void main_shutdown( void );
    void key_down(HWND hwnd, WPARAM w);
    void main_menu(HWND hwnd, unsigned int id) ;
    void key_up(WPARAM w)  ;
    void check_keys( void )  ;
    void mouse_down(LPARAM l);
    void mouse_up(LPARAM l);
    void mouse_move(LPARAM l) ;
    void timer_message(HWND hwnd, WPARAM wparam)  ;
    void window_paint(HWND hwnd)  ;
    void memdc_setup( HWND hwnd )  ;
        
    static LRESULT CALLBACK proc_wrapper(HWND hwnd, UINT message, WPARAM w, LPARAM l);
    LRESULT CALLBACK proc(HWND hwnd, UINT message, WPARAM wparam, LPARAM lparam)      ;   
           
    int get_display_x(void) { return map_display_x; };
    int get_display_y(void) { return map_display_y; };
    
            
      
  private:
           
           
    bool mouse_is_down ;
    bool right_mouse_down ;

    bool keys[256];     // Key array, true if pressed

    int updates ;                // Number of updates since last second
    int updates_per_second ;     // Number of updates per second

    bool manual_mode ;

    int timer_val ;

    unsigned int mouse_x_start;
    unsigned int mouse_y_start;

    bool show_stats ;

    int map_display_x;
    int map_display_y;       // Current Size of the client area

    profile_class profile;

    HDC memdc;                      // Device context for offscreen drawing
    HBITMAP mem_bitmap;             // Bitmap for memdc


  public:
          
    // Brushes for drawing
    HBRUSH white_brush;
    HBRUSH black_brush;
    HBRUSH green_brush;
    HBRUSH blue_brush;
    HBRUSH red_brush;
    HBRUSH color1_brush;

    // Pens for drawing    
    HPEN white_pen;
    HPEN black_pen;
    HPEN green_pen;
    HPEN blue_pen;
    HPEN color1_pen;      

    world_class *g_world;       // All world data 
};

#endif





