/*
    Robot Controller Client

    (C) 2005-2011 Jason Hunt

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

    Contact: nulluser@gmail.com

    FILE: map.cpp
    
    Robot mapping
*/

#include <windows.h>
#include <math.h>
#include <stdio.h>

#include "types.h"
#include "map.h"
#include "main.h"
#include "robot.h"
#include "world.h"

#define MODULE_NAME "[Map] "




// x, y Offsets for map directions
const int offsets [][2] = {{1,0},  {1,1},   {0,1},  {-1,1}, 
                    {-1,0}, {-1,-1}, {0,-1}, {1,-1}}; 
       





/* Draw navigation map grid */
void world_class::world_draw_map_grid(robot_class *robot,  HDC dc )
{
    double cell_x_size = g_main->get_display_x() /(double)(MAP_GRID_X);
    double cell_y_size = g_main->get_display_y() /(double)(MAP_GRID_Y);
    
    HBRUSH br;
    HPEN pen;
   // return;
        
//    log(MODULE_NAME "Draw map\n");    
    
    //  log("m2: %x\n", robot->world_get_map(0, 0));  
    
    
    int x1 = robot->get_map_x(robot->get_x());
    int y1 = robot->get_map_y(robot->get_y());
    
    


    
    
    
        
    for (int map_x = 0; map_x < MAP_GRID_X; map_x++)
        for (int map_y = 0; map_y < MAP_GRID_Y; map_y++)
        {
            int  val = robot->world_get_map(map_x,map_y);

            bool draw_cell = true;
   
   //val |=MAP_MARKED;
   
           if (val == -1)
            {
                 draw_cell = false;
               //  br = CreateSolidBrush(RGB(255, 0, 0));
               //  pen = CreatePen(PS_SOLID, 1, RGB(255,0,0));                
            } else
            if (val & MAP_MARKED)
            {
                 br = CreateSolidBrush(RGB(255, 0, 0));
                 pen = CreatePen(PS_SOLID, 1, RGB(255,0,0));                
            }
            else        
            { 

                //draw_cell = false;
                val &= ~MAP_MARKED;
    
                val *= 5;
                if (val > 255) val = 255;

//val = 0;

                br = CreateSolidBrush(RGB(0, 0, val));
                pen = CreatePen(PS_SOLID, 1, RGB(0,0,val));

               // br = CreateSolidBrush(RGB(255, 255, 255));
               //en = CreatePen(PS_SOLID, 1, RGB(255,255,255));


            }

        if (draw_cell)
        {
            SelectObject(dc, br);
            SelectObject(dc, pen);

            int screen_x = (int)(map_x * cell_x_size);
            int screen_y = (int)(map_y * cell_y_size);            

            Rectangle(dc, screen_x, screen_y, 
                         (int)(screen_x + cell_x_size) + 1, 
                         (int)(screen_y + cell_y_size) + 1);


         //  if (world.map_grid[map_x][map_y] < 255)
          /*  {
                    char b[400];
                 sprintf(b, "%4d", world.map_grid[map_x][map_y]);
                 TextOut(dc, screen_x, screen_y, b, strlen(b));
            }*/
            DeleteObject(br);
            DeleteObject(pen);
        }
    }
    
}
/* End of world_draw_map_grid */








/* Get direction to current waypoint */
void robot_class::compute_waypoint_vector(double x, double y, point &v)
{
    v.x = 0;
    v.y = 0;

  //log(MODULE_NAME "compute_waypoint_vector\n"); 

    if (num_waypoints == 0) return;
  
   double waypoint_x;
   double waypoint_y;
   
    // find current waypoint

    waypoint_type *w = waypoints;
    waypoint_type *waypoint = NULL;

    // Find current waypoint
    while (w != NULL)
    {
        if (w->num == cur_waypoint)
           waypoint = w;
        
         w = w->next;       
    }

    // Current waypoint not found
    if (waypoint == NULL)
    {
        //write_console("waypoint not found\n");
        return;
    }

    v.x = x - waypoint->x;
    v.y = y - waypoint->y;    
}
/* End of compute_waypoint_vector */



/* Convert world x to map x */
int robot_class::get_map_x(double x)
{
    unsigned int mx = (int) ((x/world->get_x_size()) * MAP_GRID_X);
    
    if (mx < 0) mx = 0;
    if (mx > MAP_GRID_X-1) mx = MAP_GRID_X-1;
    
    return(mx); 
}
/* End of get_map_x */


/* Convert world y to map y */
int robot_class::get_map_y(double y)
{
    unsigned int my = (int) ((y/world->get_y_size()) * MAP_GRID_Y);
    
    if (my < 0) my = 0;
    if (my > MAP_GRID_Y-1) my= MAP_GRID_Y-1;
    
    return(my); 
}
/* End of get_map_y */


/* Mark  location on map */
void robot_class::world_set_map(int x, int y)
{
    if (x < 0) x = 0;
    if (x > MAP_GRID_X-1) x = MAP_GRID_X-1;

    if (y < 0) y = 0;
    if (y > MAP_GRID_Y-1) y = MAP_GRID_Y-1;
    
    map_grid[x][y] = -1;
}
/* End of world_set_map */


/* Get map value */
int robot_class::world_get_map( int x, int y)
{
    if (x < 0) return(255);
    if (y < 0) return(255);
   
    if (x > MAP_GRID_X - 1) return(255);
    if (y > MAP_GRID_Y - 1) return(255);
    
    return(map_grid[x][y] & 0x0fff);    
}
/* End of world_test_map */





/* Perform map itteration */
void robot_class::world_compute_map_pass( bool &change)
{
    int temp_map[MAP_GRID_X][MAP_GRID_Y];

    bool new_change = false;

    for (int y = 0; y < MAP_GRID_Y; y++)
    {
        for (int x = 0; x < MAP_GRID_X; x++)
        {
            int v, min = 255;

            v = world_get_map(x-1, y);
            if (v < min && v != -1) min = v;
            
            v = world_get_map(x+1, y);
            if (v < min && v != -1) min = v;

            v = world_get_map(x, y-1);
            if (v < min && v != -1) min = v;

            v = world_get_map(x, y+1);
            if (v < min && v != -1) min = v;

            if (min != 255 && map_grid[x][y] != 0 && map_grid[x][y] != -1)
            {
                temp_map[x][y] = min + 1;
            }
            else
                temp_map[x][y] =map_grid[x][y];
        }       
    }

    for (int y = 0; y < MAP_GRID_Y; y++)
        for (int x = 0; x < MAP_GRID_X; x++)
        {

            if (map_grid[x][y] != temp_map[x][y])
                new_change = true;

            map_grid[x][y] = temp_map[x][y];
            
        }
            
            
    if (!new_change) change = false;
}
/* End of world_compute_map_pass */




/* Compute navigation map */
void robot_class::world_compute_map_grid( void  )
{
    //log(MODULE_NAME "world_compute_map_grid\n");

    for (int map_x = 0; map_x < MAP_GRID_X; map_x++)
        for (int map_y = 0; map_y < MAP_GRID_Y; map_y++)
            map_grid[map_x][map_y] = 255;   
 
    impact_type *i = world->get_impacts();
 
     while (i != NULL)
     {
        world_set_map(get_map_x(i->x), get_map_y(i->y));

        world_set_map(get_map_x(i->x)+1, get_map_y(i->y));
        world_set_map(get_map_x(i->x)-1, get_map_y(i->y));
        world_set_map(get_map_x(i->x), get_map_y(i->y)+1);
        world_set_map(get_map_x(i->x), get_map_y(i->y)-1);
 
         i = i->next;
    }
 
    
 
    
 
    // Mark goal
    map_grid[get_map_x(world->get_target_x())][get_map_y(world->get_target_y())] = 0;
//    map_grid[get_map_x(target_x)][get_map_y(target_y)] = 0;
 
    bool  change = true;

    // Iterate map
    while (change)
       world_compute_map_pass(change); 
 
 
    // mark path
    
    
   // if (num_robots <= 0) return;
    
    int p_x = get_map_x(get_x());
    int p_y = get_map_y(get_y());
      
    
    int val = 255;
    
    
   int old_map[MAP_GRID_X][MAP_GRID_Y];
    

    for (int y1 = 0; y1 < MAP_GRID_Y; y1++)
        for (int x1 = 0; x1 < MAP_GRID_X; x1++)
            old_map[x1][y1] = map_grid[x1][y1];
            
    int count = 0;

    int  max_path = 255;      
  

        free_waypoints();                

                
    while(val > 0 && count++<max_path-1) 
    {
        unsigned int index = world_get_lowest_index(p_x, p_y);

        // Add direction
        p_x += offsets[index][0];
        p_y += offsets[index][1];
        
        // Range check the result
        if (p_x < 0) p_x = 0;
        if (p_x > MAP_GRID_X -1) p_x = MAP_GRID_X -1;
        
        if (p_y < 0) p_y = 0;
        if (p_y > MAP_GRID_Y -1) p_y = MAP_GRID_Y -1;
        
        
        // Not a wall?, add waypoint
        if ( (val = map_grid[p_x][p_y]) != -1)
        {
            double world_x = p_x * world->get_x_size() / (double)MAP_GRID_X;
            double world_y = p_y * world->get_y_size() / (double)MAP_GRID_Y;
            
            point nav;

            compute_map_nav(nav, world_x, world_y);
            
            double dir = atan2(nav.y, nav.x);


            map_grid[p_x][p_y] |= MAP_MARKED;



           // for (int i = 0; i < num_robots; i++)
                add_waypoint( world_x, world_y, dir);
        }                        
        
   
    }
    

    
    
    // No path found
    if (count == max_path)
    {
        world->set_target(f_rand(1, world->get_x_size()-1), f_rand(1, world->get_y_size()-1));
         return;
    }

    // Number the way points
    
    waypoint_type *w = waypoints;
    
    int cur_waypointn = 0;
    
    while(w != NULL)
    {
        w->num = cur_waypointn++;
        w = w->next;
    }

    // Set starting waypoint
    cur_waypoint = num_waypoints - 1;
    
    
    /*
    
    log("after\n");
        
    for (int i = 0; i < MAP_GRID_X; i++)
    {
    for (int j = 0; j < MAP_GRID_Y; j++)
    {
    
    log(" %2d ", map_grid[i][j] );
    
    
    }
    log("\n");
    
    }*/
    
    
    
    
    
    
    
    
    
    
    
    
    
}
/* End of  world_compute_map_grid */




/* Compute index of direction pointing to smallest neighbor */
int robot_class::world_get_lowest_index(int x, int y)
{
    int lowest_index = 8;
    int lowest_num = 255;
    int v;

    for (int i  = 0; i < 8; i++)
    {    
        v =  world_get_map(x + offsets[i][0],  
                           y + offsets[i][1]);    

        if (v <lowest_num && v != -1) 
        { 
            lowest_num = v; 
            lowest_index = i;
        }
    }
    
    return(lowest_index);
}        
/* End of world_get_lowest_index */



/* Get vector for nagivation */
void robot_class::compute_map_nav(point &nav, double x, double y)
{
    double angle = 0;

    double sum_x = 0;
    double sum_y = 0;


    // Iterate over the eight directions
    for (int i  = 0; i < 8; i++)
    {    
        int v = world_get_map(  get_map_x(x)+offsets[i][0], get_map_y(y)+offsets[i][1]);    
    
        // Not a wall?
        if (v != -1)
        {
            sum_x -= v * cos(angle);
            sum_y -= -v * sin(angle);
        }

        angle += M_PI/4.0;
    }

    nav.x = sum_x;
    nav.y = sum_y;    
}
/* End of compute_map_nav */
