/*
    Robot Controller Client

    (C) 2005-2011 Jason Hunt

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

    Contact: nulluser@gmail.com

	Simple profiler.
	
	File: profile.cpp
	
	Call profile_register at the beginning of the program
		 profile_start to begin the section
		 profile_stop to end the section
		 profile_list to list the current times
*/



#include <stdio.h>
#include <string.h>

#include "utility.h"
#include "profile.h"


#define MODULE_NAME "[Profile] "



/* Start profiling */
void profile_class::start(unsigned int id)
{
	if (id < 0 || id >= num_profile) return;
	
	profiles[id].start = get_time();
}
/* End of profile start */


/* Stop profiling */
void profile_class::stop(unsigned int id)
{
	if (id < 0 || id >= num_profile) return;

	profiles[id].time  = get_time() - profiles[id].start;
}
/* End of profile stop */


/* Display profiles */
void profile_class::list( void )
{
	log(MODULE_NAME "Profile:\n");
	
	for (int i = 0; i < num_profile; i++)
		log(MODULE_NAME "  %s  -  %f seconds\n", profiles[i].name, profiles[i].time);
}
/* End of profile_list */


/* Register a new item */
void profile_class::load(char *name, unsigned int *id)
{
    log(MODULE_NAME "Profile Register: %s\n", name);

	if (num_profile == MAX_PROFILE) return;
	
	strncpy(profiles[num_profile].name, name, MAX_PROFILE_NAME);
	
	*id = num_profile++;
}
/* End of profile_register */











