/*
    Robot Controller Client

    (C) 2005-2011 Jason Hunt

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

    Contact: nulluser@gmail.com

    FILE: robot.cpp
    
    Robot methods
*/


#include <math.h>
#include <windows.h>
#include <stdio.h>


#include "types.h"
#include "utility.h"
#include "graphics.h"
#include "automode.h"
#include "robot.h"
#include "world.h"

#include "main.h"
#include "map.h"
#include "collision.h"

#include "robot_win.h"


#define MODULE_NAME "[Robot] "




/* Robot constructor */
robot_class::robot_class( world_class *w, char *v_name, 
                          char *v_host_name, int v_base_port, int v_display)

{
    world = w;


    // Copy over constructor data
    base_port = v_base_port;
    strcpy(name, v_name);
    strcpy(host_name, v_host_name);
    display = v_display;


    // Create main window
    robot_win = new robot_win_class(this);



    


    speed = 0;
    ddir = 0;


    motion(0,0);

    // Control data
//    l_motor = 0x80; 
//    r_motor = 0x80; 
//    pan = 0x80; 
//    tilt = 0x80;

    // Telemetry
    z_angle = 0;
    x_vel = 0;
    y_vel = 0;
    wifi_signal = 0;   
    ir_dist = 0;
    
    
    for (int i = 0; i < AVERAGE_SIZE; i++)
        nav_average[i] = 0;

     rev_bias = 0;
    turn_bias = 0;
    first_run = 1;

    auto_mode = 0;
 
 
     // No waypoints
    waypoints = NULL;
    num_waypoints = 0;
    cur_waypoint = 0;
 
 
 
 //    x = f_rand(x_size, world->get_x_size() - x_size);
//    y = f_rand(y_size, world->get_y_size() - y_size);
    
    
     size = 0.5;
   
    x_size = size;
    y_size = size;
    
    num_vision_points = 5;
    num_touch_points = 6;
      
    sensor_dist = (x_size + y_size) / 4.0;
    sensor_size = (x_size + y_size) / 8.0;    
                                

    x = f_rand(x_size, world->get_x_size() - x_size);
    y = f_rand(y_size, world->get_y_size() - y_size);

    dir =  f_rand(0, 2*M_PI);


                                
    // Clear touch sensor inputs
    for (int i = 0; i < 6; i++)    
        touch[i] = 0;

    // Clear vision inputs    
  /*  for (int i = 0; i < num_vision_points; i++)    
    {
        // Clear vision vectors
        end_x[2*i] = x;
        end_y[2*i] = y;
        
        end_x[2*i+1] = x;
        end_y[2*i+1] = y;
    }  */
    

    for (int i = 0; i < 6; i++)
       wall_found[i] = false;
       
    draw_vision = false;
    
    
    last_update = get_time();

    
    
    update();
    
    
 
}
/* End of robot constructor */



/* Robot destructor */
robot_class::~robot_class( void )
{
    delete robot_win;

}
/* End of robot destructor */






/* Get control data for robot */
int robot_class ::get_control_data(unsigned char *buff)
{
    unsigned char i = 0;

    if (buff != NULL)
    {
        buff[i++] = l_motor;
        buff[i++] = l_motor;
        buff[i++] = r_motor;
        buff[i++] = r_motor;
    }
        
    return i;
}
/* End of get_control_data */


/* Process robot telemetry */
void robot_class::process_telemetry(unsigned char *bf, unsigned int len)
{
    ir_dist = (float)(bf[4] + bf[5]*256)*360/65536.0;
    z_angle = (float)(bf[6] + bf[7]*256)*360/65536.0;
    x_vel = (float)(bf[8] + bf[9]*256)/65536.0   * 10;
    y_vel = (float)(bf[10] + bf[11]*256)/65536.0   * 10;
    
    
    wifi_signal = bf[14] * 100 / 255.0;
    

/*    robot_win->alert("Wifi Signal: %4.1f  Angle: %4.1f Velx: %4.1f  Vely: %4.1f",  
                        bf[14] * 100 / 255.0,
                         z_angle,
                         x_vel,
                         y_vel
                         
                         );
*/


}
/* End of process_telemetry */


/* Create telemetry string */
void robot_class::telemetry_string(char *st)
{
    sprintf((char *)st, "FPS: %d  Wifi Signal:%4.1f  IR: %4.1f Angle:%4.1f  Velx:%4.1f  Vely:%4.1f",  
                        robot_win->camera->get_fps(), wifi_signal,
                        ir_dist, z_angle, x_vel,  y_vel);
}
/* End of telemetry string */


/* Manually Set motions for the first anaimal */
void robot_class::motion(double s, double d)
{
        log(MODULE_NAME "motion %4.1f  %4.1f\n", s, d);
    
    

    if (s < -1) s = -1;
    if (s >  1) s =  1;
    
    if (d < -1) d = -1;
    if (d >  1) d =  1;
    

   
    speed = s;// * 0.02;
    ddir = d;// * 0.04;
    

    double amp = 30;


        double rm =  (speed - ddir/2);
        double lm =  (speed + ddir/2);
        
        double lm_temp = 128 + lm * amp;
        double rm_temp = 128 - rm * amp;
        
//      robot_win->alert("rm: %4.3f   lm: %4.3f ", rm, lm);
       
    //  robot_win->alert("rm_temp: %4.3f   lm_temp %4.3f ", lm * amp , rm * amp );
        
        
        
        if (lm_temp < 0) lm_temp = 0;
        if (lm_temp > 255) lm_temp = 255;
        
        if (rm_temp < 0) rm_temp = 0;
        if (rm_temp > 255) rm_temp = 255;
        
        
        l_motor = (unsigned int)lm_temp;
        r_motor = (unsigned int)rm_temp;
                
    //  robot_win->alert("  lm_temp %4.3f rm_temp: %4.3f  ", lm_temp, rm_temp);


//    l_motor = l;
//    r_motor = r;
}
/* End of animal_motion */



/* Draw an animal to a device context */
void robot_class::draw(HDC dc)
{
    
    SelectObject(dc, world->g_main->black_pen);    
    RECT r;
    world->get_bounding_rect(x, y, size, size, r);

    SelectObject(dc, world->g_main->green_brush);

    // Draw robot body
    Ellipse(dc, r.left, r.top, r.right, r.bottom);

    SelectObject(dc, world->g_main->white_brush);

    // Get distance from center and size of the sensors
    double size = sensor_size;
    double dist = sensor_dist;

    // Draw the direction arrow
    world->draw_arrow(dc, x, y, dir, size);
 
 
 
    // Draw each sensor
    // The sensor index increases in a clockwise direction

    double angle = - M_PI / 3.0;
    double delta = 2 * M_PI / 6.0;

    for (int i = 0; i < 6; i++, angle += delta)
        world->draw_sensor(dc,  x, y, dir + angle, dist, size, 
                       get_sensor_color(touch[i]));
                       
                       
  /*  for (int i = 0; i < num_vision_points * 2; i++)
    {
        if (vision[i/ 2].food > 0)
        {

            MoveToEx(dc, get_screen_x(x),get_screen_y(y),NULL);
            LineTo(dc, get_screen_x(end_x[i]),get_screen_y(end_y[i]));
        }
    }*/
        
        
    if (draw_vision)
    // Draw vision lines
    for (int i = 0; i < 6; i++)
        if (wall_found[i])
        {
            line l;
            l.x1 = x;
            l.y1 = y;
            l.x2 = wall_points[i].x;
            l.y2 =wall_points[i].y;
        
            SelectObject(dc, world->g_main->green_pen);
    
            world->draw_line(dc, l);
        }        
                
                
        double s = (x_size + y_size ) / 2.0;
    

        double d =  1.5 * s * ( 1 - ir_dist / 10.0 );


    HPEN pen1 = CreatePen(PS_SOLID, 3, RGB(0, 0, 0));

        {
            line l;
            l.x1 = x + (s/2.0)* cos(dir);;
            l.y1 = y - (s/2.0)* sin(dir);
            l.x2 = x + (1.2*s +d)* cos(dir);
            l.y2 = y - (1.2*s +d)* sin(dir);
        
            SelectObject(dc, pen1);
    
            world->draw_line(dc, l);
        }        
                 
                               
                
           DeleteObject(pen1);     
                
                
    /// Draw nav vector                
 
    double mag = nav_mag;
    
    if (mag > 0.5) mag = 0.5;
 
    SelectObject(dc, world->g_main->blue_pen);  
    
    world->draw_arrow(dc, x, y, nav_dir, 0.1);
}
/* End of draw sensors */





/* Update all values for the aminal */
void robot_class::update(void)
{
//    double x_diff = x - world->target_x;
//    double y_diff = y - world->target_y;
    
//    double target_dist = sqrt(x_diff * x_diff + y_diff * y_diff);
    
//    if (target_dist < 0.3)
    if (world->target_dist(x, y) < pow(0.3, 2))

    {
        world->random_target();
     
    
    }
    
    
    
    world->compute_grad_point(this, x, y, nav_dir, nav_mag);


    point v;
    
    compute_waypoint_vector(x, y, v);

    // Compute distance between waypoints
    double waypoint_seperation =  size;

    double dist = sqrt(v.x * v.x + v.y * v.y);


//    char b[400];    
//    sprintf(b, "Cur waypoint: %d  Dist: %f  sep_dist: %f \n", world->cur_waypoint, dist, waypoint_seperation);
//    write_console(b);
    static double last_waypoint_time = 0;

    // Point to next waypoint if close enough
    if (dist < waypoint_seperation && cur_waypoint > 0)
    {
        last_waypoint_time = get_time();
          cur_waypoint--;
    }


    if (get_time() - last_waypoint_time  > 1.0)
        world_compute_map_grid();

    
    double d = 0.5;
//double dist ;

    // Read all sensors    
    
    // Claculate distance to tip of sensor
     dist = sensor_dist + sensor_size / 2.0;    

    // Calculate angle and delta for sensor check
    double angle = - M_PI / 3.0;
    double delta = 2 * M_PI / 6.0;

    // Check the sensors moving clockwise around the animal
    for (int i = 0; i < 6; i++, angle += delta)
    {
        touch[i] = world->check_sensor(x, y, dir + angle, dist);
        
        ir[i] =  MAX_IR_DIST - world->check_vision(x, y, dir + angle, wall_found[i], wall_points[i]);
    }


  
    //nav_dir = atan2(sum_y, sum_x);

    
    nav_off = nav_dir - dir;
    
//    if (nav_off > M_PI) nav_off -= 2*M_PI;
  
    if (nav_off > 2*M_PI) nav_off -= 2*M_PI;
    if (nav_off < 0) nav_off += 2*M_PI;
    
    if (nav_off > M_PI) nav_off = nav_off - 2*M_PI;
    
        
   // nav_off = nav_off;
  
    
/*    char b[400];
    sprintf(b, "nav mag  %f \n",  nav_off);
   console(b);*/



//    angle = - M_PI / 5.0;
//    delta = M_PI / 5.0;    

   
//    check_vision(x, y, dir, num_vision_points, vision,
//        end_x, end_y);

    if (auto_mode)
    {
        robot_auto_mode();
        
    }
 
        
//       robot_win->alert("Speed: %4.3f   dDir: %4.3f  l_motor %d r_motor: %d", speed, ddir, l_motor, r_motor);

//l_motor  = 0x80;
//r_motor= 0x80;


    // Direction change rate is the difference of drive speeds
  /*  a->ddir = a->l_drive - a->r_drive;

    // Speed is the sum of drive speeds
    a->speed = (a->l_drive + a->r_drive) / 2.0;    
*/
    // Add delta to direction



    double dt = get_time() - last_update;

    last_update += dt;


//    log("dt: %f\n", dt);



    double dir_scale = 2.0;

    dir +=  dir_scale * ddir * dt;

    if (dir > M_PI ) dir -= 2*M_PI; 
    if (dir < -M_PI ) dir += 2*M_PI; 

    double speed_scale = 0.5;

    // Calulate x and y deltas
    dx = speed_scale * speed * cos(dir) * dt;
    dy = -speed_scale * speed * sin(dir) * dt;



    double test_x = x + dx;
    double test_y = y + dy;



    point p;
    bool found;

    bool hit = false;


    unsigned int check_points = 12;



    // This does the collosion detection

    // Calculate angle and delta for sensor check
     angle = 0;
    delta =  (2 * M_PI) / (double)check_points;

   // Check the sensors moving clockwise around the animal
    for (int i = 0; i < check_points; i++, angle += delta)
    { 
    double d = world->check_vision(test_x, test_y, dir + angle, found, p);
    
        if (d < size/2.0 && found)
        {
        hit =1;
//        log("hit\n");
//            dx -= (size/2.0 - d) * cos(dir + angle);
//            dy -= -(size/2.0 - d) * sin(dir + angle);    
        } 
    }
    


    
    if (!hit)
    {
        // Add deltas to positon
        x += dx;
        y += dy;
    }


        // Check collision with outer walls        
        if (x < x_size/2)
        {            
            x = x_size/2;
          //  wall_hit++;    
        }
        
        if (x > world->get_x_size() - x_size/2)
        {
            x = world->get_x_size() - x_size/2;
          //  wall_hit++;        
        }
            
        if (y < y_size/2)
        {
            y = y_size/2;
           // wall_hit++;                    
        }
            
        if (y > world->get_y_size() - y_size/2)
        {
            y = world->get_y_size() - y_size/2;
           // wall_hit++;        
        }
 
    
    static double last_x = 0;
    static double last_y = 0;
    
    if ((last_x-x) * (last_x - x) +
        (last_y-y) * (last_y - y) > (size * size)/64.0)
    {
    last_x = x;
    last_y = y;
        
    world->add_marker(x , y );
    }    
    
    
//
    
    
    
    
    
    
}
/* End of robot update */



/* Setup an animal for the first time */
void robot_class::init( void )
{
   
}
/* End of animat_init */


/* Move robot camera pan */
void robot_class ::move_pan( int v )
{
    if (v > 0)
    {
        if (pan + v < 255) pan += v;
    }

    if (v < 0)
    {
        if (pan + v > 0) pan += v;
   
    }
}
/* End of move_pan */



/* Set robot camera pan */
void robot_class ::set_pan( int v )
{
    if (v < 0) v = 0;
    if (v > 255) v = 255;

    pan = v;

}
/* End of set_pan */


/* Set robot camera tilt */
void robot_class ::set_tilt( int v )
{
    if (v < 0) v = 0;
    if (v > 255) v = 255;

    tilt = v;
}
/* End of set tilt */



void robot_class ::toggle_auto_mode( void ) 

{ auto_mode = !auto_mode; 


if (auto_mode == 0)
    stop_motion();



};


void robot_class ::stop_motion( )
{
    motion(0, 0);
  
}







/* Add waypoint to nav system */
void robot_class::add_waypoint(double x, double y, double dir)
{
    waypoint_type *w = waypoints;
    
    waypoints = (waypoint_type *) malloc(sizeof(waypoint_type));

    waypoints->x = x;
    waypoints->y = y;
    waypoints->dir = dir;
      
    waypoints->next = w;
    
    num_waypoints++;
}
/* End of world_add_waypoint */


/* Free waypoint memory */
void robot_class::free_waypoints( void )
{
    waypoint_type *tmp;
        
    while (waypoints != NULL) 
    {
        tmp = waypoints->next;

        free(waypoints);

        waypoints = tmp; // Unlink
    }
    num_waypoints = 0;
}    
/* End of world_free_waypoints */


/* Draw nav waypoints to DC */
void robot_class::draw_waypoints( HDC dc )
{
    waypoint_type *tmp = waypoints;

    SelectObject(dc, (HBRUSH)world->g_main->color1_brush);
    SelectObject(dc, (HPEN)world->g_main->color1_pen);

    while (tmp != NULL) 
    {
        RECT r;

        if (tmp->num == cur_waypoint)
        {
           SelectObject(dc, (HBRUSH)world->g_main->red_brush);
           SelectObject(dc, (HPEN)world->g_main->black_pen);
            
            
        }else
        {
           SelectObject(dc, (HBRUSH)world->g_main->color1_brush);
           SelectObject(dc, (HPEN)world->g_main->color1_pen);
        }
        
        Ellipse(dc, world->get_screen_x(tmp->x - 0.05),
                      world->get_screen_y(tmp->y - 0.05),
                      world->get_screen_x(tmp->x  + 0.05),
                      world->get_screen_y(tmp->y  + 0.05));

        tmp = tmp->next;

    }
}
/* End of draw waypoints */




