/*
    Robot Controller Client

    (C) 2005-2011 Jason Hunt

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

    Contact: nulluser@gmail.com

    FILE: robot.h
*/


#ifndef ROBOT_H
#define ROBOT_H

#define NAME_MAX    256

#define MAX_IR_DIST 2.0

#include "types.h"
#include "robot_win.h"
#include "automode.h"

#include "world.h"

class world_class;

class robot_win_class;

class robot_class
{
  public:
  
    robot_class::robot_class( world_class *w, char *v_name, 
                              char *v_host_name, int v_base_port, 
                              int v_disp);

    ~robot_class();

    void draw(HDC dc);
    void update();
    void init();
    

    void motion(double s, double d);

    void process_telemetry(unsigned char *bf, unsigned int len);
    void telemetry_string(char *st);
    void set_host_name(char *name);
    void set_name(char *name);
    
    
    void move_pan( int v );
    void set_pan( int v );
    void set_tilt( int v );
    
    int get_pan( void ) { return pan; };
    int get_tilt( void ) { return tilt; };

    char * get_host_name (void) { return host_name; } ;
    
    int get_control_data(unsigned char *buff);
    
    double get_x( void ) { return x; };
    double get_y( void ) { return y; };
    
    
    double get_dir( void ) { return dir; };
    
    void set_x( double v ) { x = v; };
    void set_y( double v ) { y = v; };
    void set_dir(  double v ) { dir = v;  };
        
double get_x_size( void ) { return x_size; };
    double get_y_size( void ) { return y_size; };
            
        
  
    void robot_auto_mode( void );
    double timer_func(double v, unsigned int timer_num, double threshhold, double lim);      
    
    void stop_motion( );

    
    void compute_avoid ( void );

    void compute_reverse ( void );      
    void compute_forward ( void );
    
    void toggle_vision( void ) { draw_vision = !draw_vision; };

    void toggle_auto_mode( void );
    
    bool get_auto_mode( void ) { return auto_mode; };
    

    
    unsigned int get_base_port( void ) { return base_port; };


    int get_display( void) { return display; };

    void set_auto_mode ( int m) { auto_mode = m; };

    void add_waypoint(double x, double y, double dir);
    void free_waypoints( void );
    void draw_waypoints( HDC dc );
    
    void compute_waypoint_vector(double x, double y, point &v);    
    void world_compute_map_pass( bool &change);
    void world_compute_map_grid( void  );
    void compute_map_nav(point &nav, double x, double y);
    int world_get_lowest_index(int x, int y);

    int get_map_x(double x);
    int get_map_y(double y);
    void world_set_map(int x, int y);
    int world_get_map(  int x, int y);

    


  private:

    double x_size, y_size;
    double x, y, dx, dy;                // (x,y) position in world space
    double dir, ddir;                 // Facing direction
    
    double speed;
    
    double sensor_dist; // Distance from center for touch sensor
    double sensor_size; // diameter of sensor
    

    unsigned num_touch_points;
    
    // Inputs to neural network
    bool touch[6];

    double ir[6];
    unsigned int num_vision_points;

    double size;

 
    bool wall_found[6];
    point wall_points[6];
    
    bool draw_vision;
  
    double nav_dir;
    double nav_mag;
    double nav_off;


    double wifi_signal;


    unsigned int base_port;
    char host_name[NAME_MAX];
    char name[NAME_MAX];

    // motor controller values
    unsigned char l_motor , r_motor , pan , tilt ;

    float z_angle ;
    float x_vel ;
    float y_vel ;
    float ir_dist ;

    // Automode stuff
    action_type actions[NUM_ACTION];
    double nav_average[AVERAGE_SIZE];


    double rev_bias ;
    double turn_bias ;
    bool first_run ;
    double stime[NUM_TIMERS];
    double limits[NUM_TIMERS];

    bool auto_mode;             // In auto mode
    bool remote;                // In remote mode
   

    int display;
    
    public:
    // Waypoint data    
    waypoint_type *waypoints;    
    unsigned int num_waypoints;
        
    // the waypoint we are tracking toward
    int cur_waypoint;
            
    double last_update;

        
    // Stored map grid
    int map_grid[MAP_GRID_X][MAP_GRID_Y];

  public:

    robot_win_class *robot_win; // Pointer to child window
    world_class *world;         // Pointer to parent world
};


#endif













