/*
    Robot Controller Client

    (C) 2005-2011 Jason Hunt

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

    Contact: nulluser@gmail.com

    FILE: robot_camera.cpp
    
    Network control for dealing with the camera feed.
*/


#include <windows.h>
#include <stdio.h>

#pragma hdrstop

#include "types.h"

#include "robot_display.h"

#include "robot_camera.h"
#include "robot_comm.h"

#include "robot_win.h"

#include "robot.h"
#include "world.h"

#include "utility.h"
#include "profile.h"
#include "decompress.h"
#include "target.h"

#define MODULE_NAME "[Camera] "


/* Robot Camera constructor */
robot_camera_class::robot_camera_class( robot_win_class *p )
{
    robot_win = p;

    image_x  = 0;
    image_y  = 0;
  // stores image from camrea
    cam_image_size = 0;
   
    
    camera_connected = false;          // true if connected
    attempting_camera_connect = false;
    getting_camera_data = false;

    frames = 0;

    mode = WAIT_SIZE;
    cam_b_index  = 0;
    
    cam_last_frame = 0;

/*
connected = false;          // true if data connected
attempting_connect = false;
 no_connect = false;  
 
  last_update = 0;

comm_state = 0;
 b_index = 0;
packet_size = 0;


*/


// update = 0;    // used to keep track of camera sweep



    camera_buffer = (color_type*) malloc(max_image_x * max_image_y * sizeof(color_type));

    if (camera_buffer == NULL)
    {
        log(MODULE_NAME "Unable to get memory for camera_buffer\n");
        return;
    }


 compressed_image = (unsigned char*) malloc(MAX_COMPRESSED_SIZE);
 
     if ( compressed_image == NULL)
    {
        log(MODULE_NAME "Unable to get memory for  compressed_image\n");
        return;
    }
 
 cam_rx_buffer =   (unsigned char *) malloc(REC_BUFFER_SIZE * 2);

     if ( cam_rx_buffer == NULL)
    {
        log(MODULE_NAME "Unable to get memory for  cam_rx_buffer\n");
        return;
    }
last_frame = (color_type*) malloc(max_image_x * max_image_y* sizeof(color_type));

    if ( last_frame  == NULL)
    {
        log(MODULE_NAME "Unable to get memory for  last_frame \n");
        return;
    }
processed_frame = (color_type*) malloc(max_image_x * max_image_y* sizeof(color_type));

    if ( processed_frame == NULL)
    {
        log(MODULE_NAME "Unable to get memory for  processed_frame\n");
        return;
    }

    targeting = false;     // targeting enabled
    d_frame = false;       // display difference frame


    update = 0;
}
/* End of robot_camera_class::robot_camera_class */



/* Robot camera destructor */
robot_camera_class::~robot_camera_class( void )
{
    free(camera_buffer);
    free(compressed_image);
    free(cam_rx_buffer);

    free(last_frame);
    free(processed_frame);

}
/* End of robot_camera_class */


/* Start robot camera */
void robot_camera_class::start( void )
{
	log(MODULE_NAME "Start\n");
    //profile_register(MODULE_NAME "Image Process", &image_process);

}
/* End of start */



/* Stop robot camera */
void robot_camera_class::stop( void )
{
	log(MODULE_NAME "Stop\n");
		
	
}
/* End of stop */

  
    
    
        
void robot_camera_class::draw( HDC hdc )
{
    
    unsigned int r= StretchDIBits(hdc, 0, 0, robot_win->get_display_x(), robot_win->get_display_y(), 
                       0, 0, image_x,  image_y,               
                       camera_buffer,
                       &buffer_info, 
                       DIB_RGB_COLORS, SRCCOPY);
                       
    
 }
 
       
    
    
    
    
    
    
    
    
    
    
    
    
void robot_camera_class::camera_check( void )
{

    if (get_time()  - cam_last_frame > 1)
    {
        request_image();
        }
        
        
       if (get_time()  - cam_last_frame > 2)
          {
             closesocket(server_camera_socket);
             camera_connected = 0;
         }

       if (!camera_connected && !attempting_camera_connect )
            connect_to_camera_server();

        
        fps = frames;
        frames= 0;
        


}
    
    
    
    
    
/* Disconnect from server */
void robot_camera_class::camera_disconnect(void)
{
    closesocket(server_camera_socket);
    camera_connected = 0;
    robot_win->alert(MODULE_NAME "Disconnect.");
}
/* End of disconnect */


/* Send keepalive data */
void robot_camera_class::request_image( void )
{
if (!camera_connected) return;


    char data_packet[1]; // the data packet
    data_packet[0] = 0x40;

   // alert(MODULE_NAME "Request Image\n");

    if (send(server_camera_socket, data_packet, 1, 0) == SOCKET_ERROR)
    {
        robot_win->alert(MODULE_NAME "Unable to send data");
        camera_connected = 0;
    }
}
/* End of send_control_state */


/* Compute 16 bit checksum */
unsigned short get_checksum16(unsigned char *buff, unsigned int len)
{
    unsigned int i;
    short check = 0;

    for (i = 0; i < len; i++)
        check += buff[i];

    return(check);
}
/* End of get_checksum */





/* setup off screen buffer */
void robot_camera_class::setup_buffer( void )
{
    log(MODULE_NAME "robot_win_class::setup_buffer\n");

    memset(&buffer_info, 0, sizeof(buffer_info));

	buffer_info.bmiHeader.biSize = sizeof(buffer_info.bmiHeader);
	buffer_info.bmiHeader.biWidth = image_x;
	buffer_info.bmiHeader.biHeight = image_y;
	buffer_info.bmiHeader.biPlanes = 1;
	buffer_info.bmiHeader.biBitCount = 24;
	buffer_info.bmiHeader.biCompression = BI_RGB;
	buffer_info.bmiHeader.biSizeImage = image_x * image_y * 3;
	buffer_info.bmiHeader.biClrUsed = 0;
	buffer_info.bmiHeader.biClrImportant = 0;
	
//	DIB_RGB_COLORS
	
	
}
/* end of setup_buffer */




/* Deal with size packet */
int robot_camera_class::size_packet(unsigned char *data)
{
    
    if (data[0] != NC_IMAGE_SIZE)
    {
        robot_win->alert(MODULE_NAME "Invalid image size packet (%x)", data[0]);
        return 1;
    }   
    
    // decode size data
    unsigned int image_x_size = (data[1] << 8) + data[2];
    unsigned int image_y_size = (data[3] << 8) + data[4];


    // resize buffer if needed
    if (image_x_size != image_x || image_y_size != image_y)
    {
        image_x = image_x_size;
        image_y = image_y_size;
		setup_buffer(); // resize buff for new image size
	}

	// decode compressed size
	c_size =  data[5] << 24;
	c_size += data[6] << 16;
	c_size += data[7] << 8;
	c_size += data[8];
            
    c_checksum = (data[9] << 8) + data[10];

    
            
            
    if (c_size == 0 || c_size > MAX_COMPRESSED_SIZE)
    {
        robot_win->alert(MODULE_NAME "Invalid image size (%d)", c_size);
        return 1;
   }
            
            
            
	c_index = 0;
	c_remain = c_size;
	
	getting_camera_data = true;
                
	log(MODULE_NAME "Size packet done (%d,%d)  len: %d\n", image_x, image_y, c_size);
	
	
	//profile_start(image_process);

	return 0;
}
/* End of size_packet */


/* Deal with image packet */
void robot_camera_class::image_packet( void )
{
    	log(MODULE_NAME "Image packet\n");
    //	show_buffer("Image:", compressed_image, c_size);
    //	getting_camera_data = false;
    
    
    unsigned short checksum = get_checksum16(compressed_image, c_size);

    request_image();


    if (checksum!= c_checksum)
    {
        robot_win->alert(MODULE_NAME "Image checksum fail. expected: %d  actual: %d  (%d)\n", 
        c_checksum, checksum,  checksum -  c_checksum);
        return;
    }
    
 	//log("Checksum Verify\n");
 	
    
   // log("Checksum Verify size: %d image_x: %d image_y: %d", c_size, image_x, image_y);
 	    
    
    //	log("Start decomp\n");
    
    
//    compressed_image[c_size-1] = 0;
      

    
    
	// decompress image into display buffer
    decompress_image(compressed_image, &c_size, camera_buffer, image_x, image_y);
                 
 	//log("Decompress ok\n");


    //   process_image(compressed_image, &c_size);
                  
          
    //  memcpy(camera_buffer, compressed_image, c_size);
                 

    //	log("Decode Complete %d\n", c_size);

/*
    FILE *f = fopen("t.jpg", "wb");
    fwrite(camera_buffer, c_size, 1, f);
    fclose(f);
*/                      
                                      
                 
  // show_buffer("cam buffer", (unsigned char *)compressed_image, c_size);
                 
                 
     cam_image_size = c_size;
                 
                 
                 
//	profile_stop(image_process);
//	profile_list();
                 
                 
                 
//	log("End decomp\n");                 
                 
	// process image if targeting is enabled
	if (targeting)  process_frame( );                 
                 
     cam_last_frame = get_time();
                 
                 
	frames++;   // increment frame counter

	// force display
	InvalidateRect (robot_win->get_hwnd(), NULL, FALSE);
                
 	log("Returning\n");
                
                

}
/* End with image packet */



int robot_camera_class::camera_process_packet(unsigned char *data, unsigned int len)
{
  //  log("Process packet. Len: %d  %x  Mode: %d\n", len, data[0], mode);
    
    
    // We are waiting for the size packet 
    if (mode == WAIT_SIZE)
    {
        if (len < SIZE_LEN) 
			return(0);
    
         if (!size_packet(data))
            mode = WAIT_DATA;
     
    
      // log("End Process packet.  Mode: %d\n",  mode);
 
        return(SIZE_LEN);
        
    }   else
    if (mode == WAIT_DATA)
    {
        unsigned int copied = 0;
        
        copied = len;
        
        if (copied > c_remain)
            copied = c_remain;
        
        c_remain -= copied;
       
       
       //	show_buffer("Process Data", data, copied);
       
        memcpy(compressed_image + c_index, data, copied);
 
        c_index += copied;
 
        if (c_remain == 0)
        {
            image_packet();
            mode = WAIT_SIZE;
        }
            
	//	log("End Process packet.  Mode: %d\n",  mode);
    
        return(copied);
        
    }   
    
    return(0);
}






/* Get data from the image server */
void robot_camera_class::read_camera_data(void)
{
     
    unsigned int n_bytes = recv(server_camera_socket, 
                                (char *)(cam_rx_buffer + cam_b_index), 
                                REC_BUFFER_SIZE - cam_b_index, 0);

    cam_b_index +=  n_bytes;

    unsigned int  processed;

    while((processed = camera_process_packet(cam_rx_buffer,  cam_b_index)) > 0)
    {
        cam_b_index -= processed;

        if (cam_b_index > 0)
    //        memmove(cam_rx_buffer, cam_rx_buffer + processed, REC_BUFFER_SIZE  - processed);
            memmove(cam_rx_buffer, cam_rx_buffer + processed, n_bytes - processed);
    }

}
/* End of read_camera_data */


/* deal with network messages */
void robot_camera_class::camera_socket_message(WPARAM wParam,LPARAM lParam)
{
    switch (WSAGETSELECTEVENT(lParam))
    {
        case FD_ACCEPT:
            break;

        case FD_CONNECT:
            if(WSAGETSELECTERROR(lParam))
            {
                robot_win->alert(MODULE_NAME "Time out.");
                attempting_camera_connect = false;
                closesocket(server_camera_socket);
            }
            else
            {
                robot_win->alert(MODULE_NAME "Connected.");
                camera_connected = true;
                attempting_camera_connect = false;
                request_image(); // start the request chain
            }
            break;

        case FD_CLOSE:
            robot_win->alert(MODULE_NAME "Server disconnect.");
            camera_connected = false;
            closesocket(server_camera_socket);
            break;

        // data back from server
        case FD_READ:
            read_camera_data();
            break;

        case FD_WRITE:
            break;
    }
}
/* end of network_message */



/* attept to connect to te server */
void robot_camera_class::connect_to_camera_server( void )
{
    char out[256];
    sprintf(out, MODULE_NAME "Connecting to %s:%d", 
                                robot_win->robot->get_host_name(),
                                robot_win->robot->get_base_port() + 1);
                                
                                
                                
    robot_win->alert(out);

    // create the socket
    server_camera_socket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);

    if (server_camera_socket == INVALID_SOCKET)
    {
        robot_win->alert(MODULE_NAME "Socket error.");
        return;
    }

    LPHOSTENT hostent;
    hostent = gethostbyname(robot_win->robot->get_host_name());

    if (!hostent)
    {
        robot_win->alert(MODULE_NAME "Host name error.");
        return;
    }

    SOCKADDR_IN server;

    memset(&server, 0, sizeof(server));

    server.sin_addr = *((LPIN_ADDR)*hostent->h_addr_list);
    server.sin_port = htons(robot_win->robot->get_base_port() + 1);
    server.sin_family = AF_INET;

    WSAAsyncSelect(server_camera_socket, robot_win->get_hwnd(),
                   WM_CAMEVENT, FD_WRITE | FD_CONNECT | FD_READ | FD_CLOSE);

    camera_connected = false;
    attempting_camera_connect = true;

    connect(server_camera_socket, (LPSOCKADDR)&server, sizeof(struct sockaddr));
}
/* End of connect_to_server */


/* Draw a box */
void robot_camera_class::draw_box(unsigned int x1, unsigned int y1, 
              unsigned int x2, unsigned int y2, 
              color_type *m)
{
    // range check
    if (x1 < 0) x1 = 0;
    if (x1 > image_x-1) x1 = image_x - 1;

    if (x2 < 0) x2 = 0;
    if (x2 > image_x-1) x2 = image_x - 1;

    if (y1 < 0) y1 = 0;
    if (y1 > image_y-1) y1 = image_y - 1;

    if (y2 < 0) y2 = 0;
    if (y2 > image_y-1) y2 = image_y - 1;
            
    for (unsigned int y = y1; y < y2 ; y++) 
    {
        unsigned int index = x1 + y * image_x;

        for (unsigned int x = x1; x < x2 ; x++)
        {
            m[index].r = 255;
            m[index].g = 0;
            m[index].b = 0;
            index++;
        }
    }
}
/* end of draw box */









