/*
    Robot Controller Client

    (C) 2005-2011 Jason Hunt

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

    Contact: nulluser@gmail.com

    FILE: robot_camera.h
*/


#ifndef ROBOT_CAMERA_H
#define ROBOT_CAMERA_H

#include "robot.h"

#define WM_CAMEVENT WM_USER+2

#define MAX_COMPRESSED_SIZE 640*480*3
#define REC_BUFFER_SIZE     1024

#define WAIT_SIZE   0
#define WAIT_DATA   1

#define SIZE_LEN 11
#define NC_IMAGE_SIZE   0x40



class robot_camera_class
{
  public:
    robot_camera_class(robot_win_class *p);
    ~robot_camera_class();

    void start( void );
    void stop( void );

    void draw( HDC hdc );

    void camera_check( void );
    void camera_disconnect(void);
    void request_image( void );
    int size_packet(unsigned char *data);
    void image_packet( void );
    int camera_process_packet(unsigned char *data, unsigned int len);
    void read_camera_data(void);
    void camera_socket_message(WPARAM wParam,LPARAM lParam);
    void connect_to_camera_server( void );
    void camera_start( void );

    void setup_buffer( void );


    // Draw box on frame
    void draw_box(unsigned int x1, unsigned int y1, 
                  unsigned int x2, unsigned int y2, 
                  color_type *m);
              
    // Target tracking
    void process_frame( void );
    unsigned char get_change(unsigned int x, unsigned int y);


    int get_fps( void ) { return fps; }
    int get_connected( void ) { return camera_connected; }

      // buffer for image processing
    color_type *last_frame;
    color_type *processed_frame;

    bool get_tergeting( void ) { return targeting; }
    bool set_tergeting( bool t ) { targeting = t; }
            
  private:
              
    unsigned int image_x;
    unsigned int image_y;


    color_type *camera_buffer;                  // stores image from camrea

    unsigned int cam_image_size;

    BITMAPINFO buffer_info;

    SOCKET server_camera_socket;

    bool camera_connected ;                     // true if connected
    bool attempting_camera_connect ;
    bool getting_camera_data ;

    unsigned int frames ;
    
    unsigned int fps ;
        
    unsigned char *compressed_image;//
    unsigned char *cam_rx_buffer;

    unsigned int mode ;                                     // COmm mode
    unsigned int c_index, c_remain, c_size;                 // Comm index, packet remaining, size
    unsigned short c_checksum;                              // packet checksum
    
    unsigned int image_process;                         

    unsigned int cam_b_index ;                              // cammera buffer index
 
    double cam_last_frame ;
    
    bool targeting;                                         // targeting enabled
    bool d_frame;                                           // display difference frame

    unsigned int update;    // used to keep track of camera sweep
   
    robot_win_class *robot_win;
};


#endif






