/*
    Robot Controller Client

    (C) 2005-2011 Jason Hunt

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

    Contact: nulluser@gmail.com

    FILE: robot_comm.cpp
    
    Robot communications
*/

#include <windows.h>
#include <stdio.h>

#pragma hdrstop

#include "..\version.h"
#include "..\netcomm.h"

#include "types.h"
#include "utility.h"
#include "robot.h"

#define MODULE_NAME  "[Nodecomm] "

/************************
*** Network Functions ***
************************/


/* Robot comm constructor */
robot_comm_class::robot_comm_class( robot_win_class *p )
{
    robot_win = p;

    connected = false;          // true if data connected
    attempting_connect = false;
    no_connect = false;         // true if we do not want to connect

    // motor controller values

    last_update = 0;

    comm_state = 0;
    b_index = 0;
    packet_size = 0;

}
/* End of Robot comm constructor */    
  

             
/* Robot comm destructor */   
robot_comm_class::~robot_comm_class(  )
{

}
/* Robot comm destructor */ 


/* Comm network error */
void robot_comm_class::network_error ( void )
{
    log(MODULE_NAME "robot_comm_class::network_error \n");

    closesocket(server_control_socket);
    connected = 0;
}
/* End of network error */


/* Comm network check */
void robot_comm_class::network_check( void )
{
/*
    if (get_time()  - last_update > 1)
    {
        request_image();
        }
        
        
       if (get_time()  - last_update > 2)
          {
             closesocket(server_control_socket);
             connected = 0;
         }
*/

    
        {
            // try to reconnect
            if (!connected && !attempting_connect && !no_connect)
                connect_to_server();
        }

       send_controller_values();
 
}
/* End of comm network check */


/* Configure and init the network */
void robot_comm_class::startup_network(HWND hwnd)
{
    // fire up winsock

}
/* End of startup network */



/* Attempt to connect to te server */
void robot_comm_class::connect_to_server( void )
{
//    char out[256];
    robot_win->alert(MODULE_NAME "Connecting to %s:%d ", 
                        robot_win->robot->get_host_name(),
                        robot_win->robot->get_base_port());

//    alert(out);

    // create the socket
    server_control_socket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);

    if (server_control_socket == INVALID_SOCKET)
    {
        robot_win->alert(MODULE_NAME "Socket error.");
        return;
    }

    LPHOSTENT hostent;

    hostent = gethostbyname(robot_win->robot->get_host_name());

    if (!hostent)
    {
        robot_win->alert(MODULE_NAME "Host name error.");
        return;
    }

    /* setup address information */
    SOCKADDR_IN server;
    
    memset(&server, 0, sizeof(server));

    server.sin_addr=*((LPIN_ADDR)*hostent->h_addr_list);
    server.sin_port=htons(robot_win->robot->get_base_port());
    server.sin_family=AF_INET;

    WSAAsyncSelect(server_control_socket, robot_win->get_hwnd(), 
                WM_NETEVENT, FD_WRITE | FD_CONNECT | FD_READ | FD_CLOSE);
    connected = false;
    attempting_connect = true;

    connect(server_control_socket,(LPSOCKADDR)&server, sizeof(struct sockaddr));
}
/* End of connect_to_server */


/* Shutdown win sock */
void robot_comm_class::shutdown_network(void)
{
    closesocket(server_control_socket);     // close server socket
    robot_win->camera->camera_disconnect();


}
/* End of shutdown_network */


/* Set controller to a contol set */
void robot_comm_class::echo_controller_value(unsigned char *comm)
{
    char out[1024];

    sprintf(out, "[Control] Return: %X %X %X %X",
                (unsigned char) comm[1], (unsigned char) comm[2],
                (unsigned char) comm[3], (unsigned char) comm[4]);

    robot_win->alert(out);
}
/* End of set controller */



/* Validate a packet */
int robot_comm_class::validate_packet( unsigned char *bf, unsigned int len )
{
    int c = get_checksum(bf, len-1);
    
    if (c != bf[len-1])
    {
        robot_win->alert(MODULE_NAME "Invalid packet checksum actual: %02x  Expected: %02x\n", bf[len-1], c);
        //        show_buffer("Buffer: ", buff, len, 0);
        
        return(1);
    }
    
    return(0);
}
/* End of validate packet */



/* Process a serial packet */
int robot_comm_class::process_packet( unsigned char *bf, unsigned int len )
{
   // if(serial.log)
//        log_packet("Process: ", bf, len);

     unsigned char packet_type = bf[0];

 //   alert(MODULE_NAME "Data for type: %d", packet_type);
        

    if (packet_type == COMM_NC_TEL )
    {
        robot_win->robot->process_telemetry(bf+1, len-1);
      // show_buffer("Telemetry: ", bf, len);
    }
    
    return(0);
}
/* End of process_packet */


    
/* Process a serial byte */    
int robot_comm_class::node_process_byte( unsigned char b )
{
    //alert("%2x \n", b);


/*    // Ignore byte in flush mode
    if (serial.flush != 0)
    {
        log_write("Flush Ignore: %x\n", b);
        return(1);
    }*/



    // Decode current state
    if (comm_state  == COMM_PACKET_WAIT)      // Waiting for packet start
    {
        b_index = 0;

        packet_size = b;
     
        rx_buff[b_index++] = b;
        
        if (packet_size < COMM_RX_MIN  )
        {
            robot_win->alert(MODULE_NAME "Invalid packet size (%d) \n", packet_size);
        
            comm_state = COMM_PACKET_WAIT;   
            
            //flush = FLUSH_PERIOD;
           
            
            return(1);                 
        }
        comm_state  = COMM_PACKET_DATA;                    
    
    } else

    if (comm_state  == COMM_PACKET_DATA)      // Getting packet data
    {
        if (b_index >= COMM_RX_MAX)
        {
            robot_win->alert(MODULE_NAME "Process byte buffer size exceeded\n");

            comm_state  = COMM_PACKET_WAIT;
            
            //serial.flush = FLUSH_PERIOD;

            return(1);
        }
    
        rx_buff[b_index++] = b;
    
        // We have the entire packer, process it
        if (b_index == packet_size)
        {
            comm_state  = COMM_PACKET_WAIT;    

            if (validate_packet(rx_buff, packet_size))
            {
                // Force the controller to timeout and flush remaining
                robot_win->alert(MODULE_NAME "Unable to validate packet\n");

                show_buffer("Data: ", rx_buff, packet_size);

               // serial_flush();
      

                return(1);
            }
    
            // Strip Packet container and process
            process_packet( rx_buff+1, packet_size - 2);
        }
    }
    return(0);
}

/* End of process_byte */



/* Get date from server and parse*/
void robot_comm_class::read_server_data(void)
{
//    unsigned char packet[5];

    unsigned char buffer[COMM_BUFF_SIZE ]; // receive buffer
    
    
    
    unsigned int n_bytes = recv(server_control_socket, (char*)buffer, COMM_BUFF_SIZE , 0);

    for (int i = 0; i < n_bytes; i++)
        node_process_byte(buffer[i]);
        

/*

   alert("Read server data: %d", n_bytes);


    unsigned int bi = 0;
    
    if (n_bytes == 0) return;
     
     
     show_buffer("Data N->C" , buffer, n_bytes);
     
     
    while(bi < n_bytes)
    {
        if (buffer[bi] == 0x20) // server wants to send a string
        {
            alert("[Control] Message from server: ");            
            bi++;
            int len = buffer[bi++]; // get string length
        
            char out[300];
            memcpy(out, buffer + bi, len);
            bi += len; // eat the string
                        
            out[len] = 0;
                        
            alert(out);            

        }  else
        if (buffer[bi] == 0x30) // echo controller value
        {
             echo_controller_value(buffer);
             bi+=5;
        }   else         
           bi++; // ignore
           
           
           
           
    }*/
}
/* End of read server data */


/* End of read_client_data */
void robot_comm_class::send_string( char *str )
{
    unsigned int len = strlen(str);
    
    if (len > 255) len = 255;
    
    robot_win->alert(MODULE_NAME "Send string: %s", str);
    
    
    char packet[len+2];
    
    packet[0] = 0x20; // display string
    packet[1] = len; // length
    
    memcpy(packet+2, str, len); 

    // note, the string is not null terminated
    
    if (send(server_control_socket, packet, strlen(packet) + 1, 0) == SOCKET_ERROR)
    {
        robot_win->alert(MODULE_NAME "Unable to send data");
        
        network_error ();
        
        
    }
}
/* End of send string */


/* Sends version string to client */
void robot_comm_class::send_version( void )
{
    send_string(VERSION);
}
/* End of send_version */


/* Send a control packet to the server */
void robot_comm_class::send_packet(unsigned char *packet, unsigned int size)
{
    if (send(server_control_socket, (char *)packet, size, 0) == SOCKET_ERROR)
    {
        robot_win->alert(MODULE_NAME "Unable to send data");

        char out[256];
        sprintf(out, "Error: %ld ", WSAGetLastError());
        robot_win->alert(out);
    
        network_error ();
    }
}
/* End of set_packet */


/* egnage firing system */
void robot_comm_class::fire_on(void)
{
    unsigned char p[1]; 
    
    p[0] = 0x40;
        
    send_packet(p, 1);
   // alert("Fire on");    
}
/* End of fire_on */


/* Disengage firing system */
void robot_comm_class::fire_off(void)
{
    unsigned char p[1]; 
    
    p[0] = 0x41;

    send_packet(p, 1);

   // alert("Fire on");    
}
/* End of fire off */


/* Send the controller values to the server */
void robot_comm_class::send_controller_values()
{
    if (!connected) return;


    unsigned char control_data[100];

    // get control buffer,  Leave room for packet type
    unsigned int data_size =  robot_win->robot->get_control_data(control_data+1); 
    

    char out[256];
    sprintf(out, "[Control] Send: %X %X %X %X", control_data[1],
                                                control_data[2],
                                                control_data[3],
                                                control_data[4]);
//    robot_win-> alert(out);


    control_data[0] = 0x30; // command byte data (01 == update movement data)

    send_packet(control_data, 5);
}
/* End of send_controller_values */


/* Force stop */
void robot_comm_class::motion_stop( void )
{
    robot_win->alert(MODULE_NAME "Stopping");

//    robot_win->robot->l_motor = 0x80;
//    robot_win->robot->r_motor = 0x80;

    robot_win->robot->motion(0, 0);
   
    send_controller_values();
}
/* End of force stop */


/* Send keepalive data */
void robot_comm_class::send_keep_alive( void )
{
/*    char data_packet[5]; // the data packet

    memset(data_packet, 0, 5); // make the packet all zeros

    send_packet(data_packet, 5);*/
    
    
    send_controller_values();
}
/* End of send_control_state */


/* Send keepalive data */
void robot_comm_class::remote_shutdown( void )
{
    unsigned char data_packet[5]; // the data packet

    memset(data_packet, 0, 5); // make the packet all zeros

    data_packet[0] = 0x02;

    send_packet(data_packet, 5);

    robot_win->alert(MODULE_NAME "Sent remote shutdown command.");
}
/* End of send_control_state */


/* Connect to server */
void robot_comm_class::do_connect(void)
{
//    closesocket(server_control_socket);
    connected = false;
    no_connect = false;
}
/* End of conenct to server */


/* Disconnect from server */
void robot_comm_class::disconnect(void)
{
    send_string("Good bye!");
    closesocket(server_control_socket);
    connected = false;
    no_connect = true;
    robot_win->alert(MODULE_NAME "Disconnect.");
    
    robot_win->camera->camera_disconnect();
}
/* End of disconnect */


/* Get a hostname from the user */
void robot_comm_class::get_servername( void )
{
/*    char tmp_name[128];

    HWND hdialog;

    if (GetDlgItemText(NULL, ID_ITEMNAME, tmp_name, 128))
        set_host_name(tmp_name);*/
}
/* End of get hostname */


/* Deal with network messages */
void robot_comm_class::control_socket_message(WPARAM wParam,LPARAM lParam)
{
    switch (WSAGETSELECTEVENT(lParam))
    {
        case FD_ACCEPT:
            break;

        case FD_CONNECT:
            if(WSAGETSELECTERROR(lParam))
            {
                robot_win->alert(MODULE_NAME "Time out.");
                attempting_connect = 0;
                closesocket(server_control_socket);
            }
            else
            {
                robot_win->alert(MODULE_NAME "Connected.");
                connected = 1;
                attempting_connect = 0;
                send_version();
            }
            break;

        case FD_CLOSE:
            robot_win->alert(MODULE_NAME "Server disconnect.");
            connected = 0;
            closesocket(server_control_socket);
            break;

        // data back from server
        case FD_READ:
            read_server_data();
            break;

        case FD_WRITE:
            break;
    }
}
/* End of network_message */


/*******************************
*** End of Network Functions ***
*******************************/


