/*
    Robot Controller Client

    (C) 2005-2011 Jason Hunt

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

    Contact: nulluser@gmail.com

    FILE: robot_display.cpp
    
    Robot window display
*/


#include <windows.h>
#include <wingdi.h>
#include <stdio.h>


#pragma hdrstop


#include "types.h"

#include "utility.h"
#include "profile.h"

#include "robot_camera.h"
#include "robot_display.h"


#include "robot.h"

#define MODULE_NAME "[Display] "





/* Eeset the text area */
void robot_win_class::clear_text_area( void )
{
    log(MODULE_NAME "robot_win_class::clear_text_area\n");

    memset(text_area, ' ', text_y_size * text_x_size);
}
/* End fo clear text area */


/* Adds a line to the text buffer */
void robot_win_class::do_add_line(char *line)
{
 //   log(MODULE_NAME " robot_win_class::do_add_line\n");

    // scrool the display up
    if (text_y == text_y_size)
    {
        for (int i = 1; i < text_y_size; i++)
            memcpy(text_area[i-1], text_area[i], text_x_size);
    }
    else
        text_y ++;

    // erase the line
    memset(text_area[text_y-1], ' ', text_x_size);

    // copy the string
    unsigned int line_len = strlen(line);

    if (line_len > text_x_size)
        line_len = text_x_size;

    memcpy(text_area[text_y-1], line, line_len);

    // force paint
    if (main_win != NULL)
        InvalidateRect (main_win, NULL, FALSE);
        
}
/* End of add line */


/* Write formatted data to the debugging console */
void robot_win_class::alert(char *fmt, ... )
{
   // log(MODULE_NAME "robot_win_class::alert\n");
  
    const unsigned int max = 1024;

    va_list args;
    char text[max];

    char out_str[max] = {0};
    //    do_add_line("ss");
    

    va_start(args, fmt);
    vsnprintf(text, max, fmt, args);

    
   do_add_line(text);
  //  DWORD out = 0;
    
//    WriteConsole(h_console, text, strlen(text), &out, NULL);

    va_end(args);    
    
   //     do_add_line(out_str);
    
    log("%s\n", text);
 
}
/* End of write_console */


/* Show the text buffer on the screen */
void robot_win_class::paint_main(void)
{
  //  log(MODULE_NAME "robot_win_class::paint_main\n");

    PAINTSTRUCT ps;
    HDC hdc = BeginPaint(main_win,&ps);


    SetStretchBltMode(hdc, HALFTONE);

    SetBrushOrgEx(hdc, 0, 0, NULL);



    static unsigned int reg = 0;
    static unsigned int p_id = 0;
    
    if (reg == 0)
    {
        reg = 0;
//        robot->world->g_main->profile.load("Blit", &p_id);
    }
    
//    profile_start(p_id);
   
    // copy camera buffer
    
    camera->draw(hdc);
    

                
//    profile_stop(p_id);
                           

    if (display_text)
    {
        SelectObject(hdc, GetStockObject(SYSTEM_FIXED_FONT));

        // draw text lines
        for (unsigned int r = 0; r < text_y_size; r++)
            TextOut(hdc, 0, r * font_y_size + display_y, text_area[r], text_x_size);
    }



    char ts[1000];

    robot->telemetry_string(ts);
    
    TextOut(hdc, 0, 0, ts, strlen(ts));





    DeleteDC(hdc);

    EndPaint(main_win, &ps);
}
/* End of display_text_area */


/* Called when we resize */
void robot_win_class::resize_window(unsigned int x_size, unsigned int y_size)
{
    log(MODULE_NAME "robot_win_class::resize_window\n");
    display_x  = x_size;

    if (display_text)
        display_y  = y_size - text_area_height;
    else
        display_y  = y_size;
}
/* End of resize */


