/*
    Robot Controller Client

    (C) 2005-2011 Jason Hunt

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

    Contact: nulluser@gmail.com

    FILE: robot_input.cpp
    
    Robot window input
*/


#include <windows.h>

#pragma hdrstop

#include "utility.h"
#include "robot.h"
#include "target.h"

#define MODULE_NAME "[Client Input] "


/* Inital key setup */
void robot_win_class::reset_keys( void )
{
    log(MODULE_NAME "robot_win_class::reset_keys\n");

    for (unsigned int i = 0; i < 256; i++) 
        keys[i] = false;

    update_movement();

    speed = 8;   
}
/* End of reset_keys */


/* Deal with changed key values */
void robot_win_class::update_movement( void  )
{
    log(MODULE_NAME " robot_win_class::update_movement\n");


      if (keys[VK_RIGHT])    robot->move_pan(-3);
      if (keys[VK_RIGHT])    robot->move_pan(3);
    
//    if (keys[VK_RIGHT] && robot->pan > 3) robot->pan -= 3;
//    if (keys[VK_LEFT] && robot->pan < 252) robot->pan += 3;


    if (keys['W']) 
    {
        if (keys['A']) robot->motion( 1, 0.5 ); else
        if (keys['D']) robot->motion( 1, -0.5 ); else
             robot->motion( 1, 0 );
    }else
    if (keys['S']) 
    {
        if (keys['A']) robot->motion( -1, -0.5 ); else
        if (keys['D']) robot->motion( -1, 0.5 ); else
            robot->motion( -1, 0 );
    } else
    
    
    if (keys['A']) { robot->motion( 0, 1 ); } else // left spin
    if (keys['D']) { robot->motion(0, -1 ); } else // right spin
        robot->motion( 0, 0 ); 

    comm->send_controller_values();
}
/* End of update movement */


/* Called when a key is pressed */
void robot_win_class::process_key_down(unsigned char key)
{
    log(MODULE_NAME "robot_win_class::process_key_down\n");

    // toggle logging
    if (key == 'L')
    {
        set_logging(!get_logging());
        
        
        if (get_logging())    
            alert(MODULE_NAME "Logging enabled");
        else
            alert(MODULE_NAME "Logging disabled");
    } else

    // toggle rebug display
    if (key == 'P') 
    {
        display_text = !display_text;        
        // resize the window if needed
        if(display_text)
            display_y -= text_area_height;
        else
            display_y += text_area_height;
    } else

    // escape key
    if (key == VK_ESCAPE) 
    {
        /*camera->camera_disconnect();
        comm->shutdown_network();
        PostQuitMessage(0);*/
    } else

//    if (key == VK_SPACE && !keys[VK_SPACE])  fire_on(); else
//    if (key == 'V') d_frame = !d_frame; else
    if (key == 'T') 
    {
        if (camera->get_tergeting())
        {
            alert(MODULE_NAME "Targeting enabled");
            camera->set_tergeting(1);
        }
        else
        {
            camera->set_tergeting(0);
            comm->fire_off();       
            alert(MODULE_NAME "Targeting disabled");
        }            
    }    else
    
    if (key == 'M') 
    {
        camera_mouse = !camera_mouse; 

    }else
    
        if (key == VK_SPACE) 
    {

        robot->toggle_auto_mode();
        
        if (robot->get_auto_mode())
            alert(MODULE_NAME "Auto mode");
        else
            alert(MODULE_NAME "Manual mode");                
            

    }else
    
    if (key == '1') speed = 8; else
    if (key == '2') speed = 10; else
    if (key == '3') speed = 12; else
    if (key == '4') speed = 14; else
    if (key == '5') speed = 16; 

    bool ch = ((key == VK_LEFT ) || (key == VK_RIGHT) || 
               (key == 'A' && !keys['A']) || 
               (key == 'D' && !keys['D']) || 
               (key == 'S' && !keys['S']) || 
               (key == 'W' && !keys['W']) || 
               (key == 'Z' && !keys['Z']) || 
               (key == 'X' && !keys['X']) ||
               (key == 'C' && !keys['C']) || 
               (key == 'Q' && !keys['Q']) || 
               (key == 'E' && !keys['E']));

    // Mark key
    keys[key] = true;

    if (ch)
        update_movement();
}
/* End of process key down */


/* Called when a key is released */
void robot_win_class::process_key_up( unsigned char key )
{
    log(MODULE_NAME "robot_win_class::process_key_up\n");

//    if (key == VK_SPACE) fire_off();

    bool ch = key == VK_LEFT  || key == VK_RIGHT || 
              key == 'A' || key == 'D' || key == 'S' || 
              key == 'W' || key == 'Z' || key == 'X' ||
              key == 'C' || key == 'Q' || key == 'E';

    keys[key] = false;

    if (ch)
        update_movement();
}
/* End of process key up */


/* Handle mouse movement */
void robot_win_class::process_mouse( LPARAM lParam )
{
//    log(MODULE_NAME "robot_win_class::process_mouse\n");

    if (!camera_mouse) return;

    unsigned int pos_x = lParam & 0xffff; // decode position
    unsigned int pos_y = lParam >> 16; 
     
    // range check

    if (pos_x < 0) pos_x = 0;
    if (pos_x > display_x - 1) pos_x = display_x - 1;

    if (pos_y < 0) pos_y = 0;
    if (pos_y > display_y - 1) pos_y = display_y - 1;
     
    // scale into servo pos
    robot->set_pan ( 255 - (unsigned char)((255 / double(display_x)) * pos_x) );
    robot->set_tilt( 255 - (unsigned char)((255 / double(display_y)) * pos_y) );

    //if (window_active)
//        send_controller_values();
}
/* End of process_mouse */



