/*
    Robot Controller Client

    (C) 2005-2011 Jason Hunt

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

    Contact: nulluser@gmail.com

    FILE: robot_win.cpp
        
    Robot window
*/


#include <windows.h>
#include <stdio.h>

#pragma hdrstop
#include "types.h"

#include "utility.h"
#include "profile.h"
#include "..\version.h"
#include "resource.h"
#include "robot_display.h"
#include "robot_comm.h"
#include "robot_camera.h"
#include "robot_input.h"

#include "robot_win.h"

#define MODULE_NAME "[Client Main] "

#include "main.h"

#define SEND_KEEP_ALIVE_TIME 100 // time in ms to set the keep alive data


/* Robot window constructor */
robot_win_class::robot_win_class(robot_class *p)
{
    log(MODULE_NAME "robot_win_class::robot_win_class\n");

    robot = p;

    // inital display size
    display_x  = 640;
    display_y  = 480 + 80;



    display_text = true; // true if you want to see debugging iinformation


    for (int i = 0; i < text_y_size; i++)
        for (int j = 0; j < text_x_size; j++)
            text_area[i][j] = ' ';


    text_y = 0;                            // current y_pos in text buffer

  


    //bool keys[256]; // key track of keys

    speed = 8;

    main_win = NULL;            // keep track of the main window

    window_active = 0;

    camera_mouse = true; // true if the camera follows the mouse
       // true if we do not want to connect




// Get buffers







 
    camera = new  robot_camera_class(this) ;
    comm= new  robot_comm_class(this) ;




  
  
 if (setup_window(NULL, 1)) return;

//    parse_command_line(lpCmdLine); // get the command line values before we start up






}
/* End of Robot window constructor */



/* Robot window destructor */
robot_win_class::~robot_win_class( void )
{
    log(MODULE_NAME "robot_win_class::~robot_win_class\n");


    delete comm;    
    delete camera;
}
/* End of window constructor */











/* Start robot window */
void robot_win_class::system_startup(HWND hwnd)
{
    log(MODULE_NAME "robot_win_class::system_startup\n");
 
   main_win = hwnd; // save the handle

 
   // return;

       
    //setup_buffer(); // setup image buffers

    reset_keys(); // clear set keys
    
    // setup keep alive timer
    if(SetTimer(hwnd, ID_KEEP_ALIVE, SEND_KEEP_ALIVE_TIME, NULL) == 0)
        alert(MODULE_NAME "Unable to set timer");
        
    // setup second timer
    if(SetTimer(hwnd, ID_SECOND, 1000, NULL) == 0)
        alert(MODULE_NAME "Unable to set timer");

    comm->startup_network(hwnd);  // start the network
                
	camera->start();
	
//	mapping_start();
                
    alert(MODULE_NAME "Version: %s", VERSION);
}
/* End of system_startup */


/* Stop robot system */
void robot_win_class::system_shutdown(void)
{
    log(MODULE_NAME "robot_win_class::system_shutdown\n");
    
    comm->shutdown_network();     // close network connections
    PostQuitMessage(0);     // shutdown application
}
/* end of system_shutdown */


/* Deal with the system timer */
void robot_win_class::timer_message(WPARAM wParam)
{
    //log(MODULE_NAME "robot_win_class::timer_message\n");
        
    if (wParam == ID_KEEP_ALIVE)
    {
        // tell the server that we are still alive
        if (comm->get_connected())
            comm->send_keep_alive();
            InvalidateRect(main_win, NULL, 0);
            
    }

    // one second timer
    if (wParam == ID_SECOND) 
    {
        log(MODULE_NAME "Second\n");
        
      //  profile_list();
        
       if (camera->get_connected())
        {
            alert(MODULE_NAME "FPS: %d", camera->get_fps());
          
          // request_image();
        }
        
        camera->camera_check();
        
        comm->network_check();

    }
}
/* End of timer message */


/* Check window activity */
void robot_win_class::check_active(int active)
{
    log(MODULE_NAME "robot_win_class::check_active\n");


    window_active = active;
    if (window_active == 0)
    {

        reset_keys();
    
    
       // alert("deactivate window");
        comm->motion_stop();
    }// else
      //  alert("activate window");
}
/* End of check active */


/* Handles menu commands */
void robot_win_class::handle_menu( UINT menu_command )
{
    log(MODULE_NAME "robot_win_class::handle_menu\n");

    switch(menu_command)
    {
        case IDM_SERVER: comm->get_servername(); break;

        case IDM_CONNECT: comm->do_connect(); break;

        case IDM_DISCONNECT: comm->disconnect(); break;

        case IDM_STOP: comm->set_noconnect(1); break;

        case IDM_EXIT: system_shutdown(); break;

        case IDM_SHUTDOWN: comm->remote_shutdown(); break;
    }
}
/* End of hendle_menu */


/* Wrapper for the camera window procedure */
LRESULT CALLBACK robot_win_class::proc_wrapper(HWND hwnd, UINT message, WPARAM w, LPARAM l)
{
//    log(MODULE_NAME "robot_win_class::proc_wrapper\n");

    // Get object pointer from window data and call actual wndproc
        
    // Set the class pointer on creation
    if (message == WM_NCCREATE)
    {     
        if(l)
        {
            LONG handle = (LONG)((CREATESTRUCT*)l)->lpCreateParams;
        
            if(handle)
                SetWindowLong(hwnd, GWL_USERDATA, handle);
        }
    }
    
    robot_win_class *win = (robot_win_class*)::GetWindowLong(hwnd, GWL_USERDATA);
            
    // Call the member function     
    win -> proc(hwnd, message, w, l);
}
/* End of cam_proc_wrapper */



/* message handler */
LRESULT CALLBACK robot_win_class::proc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
//    log(MODULE_NAME "robot_win_class::proc\n");
    
    switch (message)                 
    {
        case WM_CREATE:     system_startup(hwnd);    break;
        case WM_NETEVENT:   comm->control_socket_message(wParam, lParam); break;
        case WM_CAMEVENT:   camera->camera_socket_message(wParam, lParam); break;
        case WM_TIMER:      timer_message(wParam); break;
        case WM_COMMAND:    handle_menu(wParam); break;
        case WM_DESTROY:    system_shutdown(); break;
        case WM_KEYDOWN:    process_key_down(wParam & 0xff); break;
        case WM_KEYUP:      process_key_up(wParam & 0xff); break;
        case WM_MOUSEMOVE:  process_mouse(lParam); break;
//        case WM_LBUTTONDOWN:  fire_on(); break;
//        case WM_LBUTTONUP:  fire_off(); break;
        
        case WM_PAINT:      paint_main(); break;
        case WM_SIZE:       resize_window(lParam & 0xffff, lParam >> 16); break;
        case WM_ACTIVATEAPP: check_active(wParam); break;
  
         case   WM_KILLFOCUS: check_active(0);  break;
  
  
        default: return DefWindowProc (hwnd, message, wParam, lParam);
    }

    return(0);
}
/* End of winproc */


/* Create the main window */
int robot_win_class::setup_window(HINSTANCE hThisInstance, int nCmdShow)
{
    log(MODULE_NAME "robot_win_class::setup_window\n");

    WNDCLASSEX wincl;  

    char class_name[] = "RobotWin"; //the class name

    wincl.hInstance = hThisInstance;
    wincl.lpszClassName = class_name;
    wincl.lpfnWndProc = proc_wrapper;      
    wincl.style = CS_DBLCLKS;         
    wincl.cbSize = sizeof (WNDCLASSEX);

    wincl.hIcon = LoadIcon (NULL, IDI_APPLICATION);
    wincl.hIconSm = LoadIcon (NULL, IDI_APPLICATION);
    wincl.hCursor = LoadCursor (NULL, IDC_ARROW);
    wincl.lpszMenuName = NULL;               
    wincl.cbClsExtra = 0;                    
    wincl.cbWndExtra = 0;                    
    wincl.hbrBackground = (HBRUSH)(COLOR_WINDOW+1); 


    // Only register the class if we have not already
    if (FindWindow(class_name, NULL) == NULL)
        if (!RegisterClassEx(&wincl)) 
        {
            log(MODULE_NAME " Unable to register window class");
            return 1;
        }


    char name[256];
    sprintf(name, "Robot View");

    HWND hwnd = CreateWindowEx (0, class_name,
                                name,  
                                WS_OVERLAPPEDWINDOW,
                                160,0,   // pos
                                display_x + 8, // x_ size
                                font_y_size * text_y_size + display_y + 44,

                               HWND_DESKTOP, NULL,
                               hThisInstance, this);

    if (!hwnd) return(1); // make sure it was made


    // load the menu
    HMENU hMenu = LoadMenu (hThisInstance, "MAINMENU");
    SetMenu (hwnd, hMenu);

    if (robot->get_display())
        ShowWindow (hwnd, nCmdShow); // display the window

    clear_text_area( );

    return(0);
}
/* End of setup window */















