/*
    Robot Controller Client

    (C) 2005-2011 Jason Hunt

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

    Contact: nulluser@gmail.com

    FILE: robot_win.h
*/

#ifndef ROBOT_WIN_H
#define ROBOT_WIN_H

#define ID_KEEP_ALIVE  1    // id for timer for safety check
#define ID_SECOND      2    // id for timer for safety check

#include "main.h"
#include "robot_display.h"
#include "robot_comm.h"
#include "robot_camera.h"
#include "robot.h"


/* Robot Window Class */
class robot_win_class 
{
  public:

    robot_win_class(robot_class *p);
    ~robot_win_class();


    void clear_keys(void);

    void clear_text_area( void );
    void do_add_line(char *line);
    void alert(char *fmt, ... );
    void paint_main(void);
    void resize_window(unsigned int x_size, unsigned int y_size);
    void reset_keys( void );

    void update_movement( void  );
    void process_key_down(unsigned char key);
    void process_key_up( unsigned char key );
    void process_mouse( LPARAM lParam );

    void system_startup(HWND hwnd);
    void system_shutdown(void);
    void timer_message(WPARAM wParam);
    void check_active(int active);
    void handle_menu( UINT menu_command );

    static  LRESULT CALLBACK robot_win_class::proc_wrapper(HWND hwnd, UINT message, WPARAM w, LPARAM l);
    LRESULT CALLBACK proc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);

    int setup_window(HINSTANCE hThisInstance, int nCmdShow);


    void draw_box(unsigned int x1, unsigned int y1, 
              unsigned int x2, unsigned int y2, 
              color_type *m);
              
              
    int get_display_x ( void ) { return display_x; };
    int get_display_y ( void ) { return display_y; };
    
    HWND get_hwnd ( void ) { return main_win; };
              
              
  private:

    // inital display size
    unsigned int display_x;
    unsigned int display_y;

    bool display_text; // true if you want to see debugging iinformation
    char text_area[text_y_size][text_x_size];  // the data for textr display
    unsigned int text_y;                            // current y_pos in text buffer

   
    bool keys[256]; // key track of keys

    int speed;

  
    unsigned int window_active;

    bool camera_mouse; // true if the camera follows the mouse
             
     
    HWND  main_win;
    
    
  public:




    robot_camera_class *camera;
    robot_comm_class *comm;

              
    robot_class *robot;
};

#endif





