/*
    Robot Controller Client

    (C) 2005-2011 Jason Hunt

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

    Contact: nulluser@gmail.com

    FILE: world.cpp
*/


#include <windows.h>
#include <math.h>
#include <stdio.h>

#include "types.h"
#include "config.h"
#include "utility.h"
#include "main.h"
#include "graphics.h"
#include "main.h"
#include "robot.h"
#include "world.h"
#include "map.h"
#include "collision.h"

#define MODULE_NAME "[World] "




world_class::world_class( main_class *m)
{
    g_main = m;


    log(MODULE_NAME "constructor\n");
 // Size in world space
    x_size = 10;
    y_size = 10;



    // Run time tweakable values
    scale1 = 0.11;
    scale2 = 0.86;
    scale3 = 0.193;
        
 
    // Start showing walls
    show_walls = true;
 



    // No walls
    walls = NULL;
    num_walls = 0;

    // No markers
    markers = NULL;
    
    
    // No Impacts
    impacts = NULL;


    show_markers = false;
    show_impacts = false;
    show_grad = false;
    show_map = false;


    
    target_enabled = true;
    target_x = x_size /2.0;
    target_y = y_size /2.0;
    
    
    // Setup surronding walls
    
    double wall_width = 0.05;
    
    // top
    add_wall(0, 0, x_size, wall_width);

    // Bottom
    add_wall(0, y_size-wall_width, x_size, wall_width);
    
    // Left
    add_wall(0, 0, wall_width, y_size);    
    
    // Right
    add_wall(x_size-wall_width, 0, wall_width, y_size);
    

    // Setup cross walls in center
   // world_add_wall(world.x_size/2 - world.x_size/4 , world.y_size/2-wall_width/2, world.x_size/2, wall_width);
   // world_add_wall(world.x_size/2-wall_width/2, world.y_size/2 - world.y_size/4 , wall_width, world.y_size/2);


    num_robots = 0;
   
    load_robots();
   
  
    // Compute first map interation
    
    for (int i = 0; i < num_robots; i++)
        robots[i]->world_compute_map_grid();
}


world_class::~world_class( void )
{
    num_walls = 0;
    
    free_walls();    
    free_markers();
    free_impacts();
    
    for (int i = 0; i < num_robots; i++)
        robots[i]->free_waypoints();

    
    for (unsigned int i = 0; i < num_robots; i++)
        delete robots[i];

}




void world_class::robot_parse(char *line)
{

//    log(MODULE_NAME "Parse: %s\n", line);

    unsigned int index = 0;
    char str[CONFIG_LINE_MAX];
    
    get_next_string(line, index, str);

    if (str[0] == 0) return;
         
    char robot_name[strlen(str)];
    
    strcpy(robot_name, str);   
   
   
    get_next_string(line, index, str);
   
    if (str[0] == 0) return;

    char robot_host_name[strlen(str)];
    
    strcpy(robot_host_name, str);   
   

//    log(MODULE_NAME "Line1: %s\n", str);
        
    get_next_string(line, index, str);
    if (str[0] == 0) return;
       
    int robot_base_port = 0;
 
    sscanf(str, "%d", &robot_base_port);
    
        
        
    get_next_string(line, index, str);
    if (str[0] == 0) return;
       
    int display = 0;
        
    sscanf(str, "%d", &display);
    

    get_next_string(line, index, str);
    if (str[0] == 0) return;
       
    int auto_mode = 0;
        
    sscanf(str, "%d", &auto_mode);
               
                
    log(MODULE_NAME "Robot. Name: %s   Host:  %s  Port:  %d\n", 
        robot_name, robot_host_name, robot_base_port);
        
        
    if (num_robots >= MAX_ROBOT-1) 
    {
        log(MODULE_NAME "Too many robots\n");
        return;
    }
    
      
    
        
    robots[num_robots] = new robot_class(this, robot_name, robot_host_name, robot_base_port, display);

    robots[num_robots]->set_auto_mode(auto_mode);


    num_robots++;
}



void world_class::robot_parse_wrapper(void *w, char *line)
{


    world_class *world = (world_class *)w;

    world->robot_parse(line);


}


void world_class::load_robots( void )
{
    load_config("robots.txt", robot_parse_wrapper, this);
   

    // Setup robot
    for (unsigned int i = 0; i < num_robots; i++)
        robots[i]->init();
}





/* Draw entire world */
void world_class::draw( HDC dc )
{

    SelectObject(dc, g_main->white_pen);
    SelectObject(dc, g_main->white_brush);
      
    // Clear entire area
    Rectangle(dc, 0, 0, g_main->get_display_x(), g_main->get_display_y());


//world_draw_grad(dc);


   if (show_grad)
        draw_grad(dc);


//
    if (num_robots > 0)
        if (show_map)
         world_draw_map_grid(robots[0], dc);


    SelectObject(dc, g_main->black_pen);
    SelectObject(dc, g_main->white_brush);

    draw_markers(dc);




    if (show_walls)
        draw_walls(dc);


    draw_target(dc, target_x, target_y);
    




    if (show_impacts)
       draw_impacts(dc);    



    for (int i = 0; i < num_robots; i++)
        robots[i]->draw_waypoints(dc);

    for (unsigned int i = 0; i < num_robots; i++)
        robots[i]->draw(dc);
    
}
/* End of world_draw */


/* Update all world data */
void world_class::update( void )
{
 //   world_compute_map_grid();

    for (unsigned int i = 0; i < num_robots; i++)
        robots[i]->update();
}
/* End of world_update */


    
/* Transform from world space to screen space */
int world_class::get_screen_x( double x )
{
    return((int)(x / (double) x_size * g_main->get_display_x()));
}
/* End of get_screen_x */


/* Transform from world space to screen space */
int world_class::get_screen_y( double y )
{
    return((int)(y / (double) y_size * g_main->get_display_y()));
}
/* End of get_screen_y */


/* Return world cords of a screen pos */
double world_class::get_world_x( unsigned int x)
{
    return((x / (double)g_main->get_display_x()) * x_size);        
}
/* End of get_world_x */

/* Return world cords of a screen pos */
double world_class::get_world_y( unsigned int y)
{
    return((y / (double)g_main->get_display_y()) * y_size);        
}
/* End of get_world_x */


/* Return a bounting box in screen space for an object */
void world_class::get_bounding_rect(double x, double y, double x_size, double y_size, RECT &r)
{
    r.left = get_screen_x(x - x_size / 2);
    r.top = get_screen_y(y - y_size / 2);
    r.right = get_screen_x(x + x_size / 2);
    r.bottom = get_screen_y(y + y_size / 2);
}
/* End of get_bounding_rect */
    


/* Return true if points is inside of wall */    
bool world_class::check_wall_point( double x, double y )
{
    wall_type *w = walls;

    while (w != NULL) 
    {
        if (x >= w->x && x <= w->x + w->x_size && 
            y >= w->y && y <= w->y + w->y_size)
            return(1);

        w = w->next;

    }
    return(false);
}    
/* End of world_check_wall_point */


/* Add a wall to the world */
void world_class::add_wall(double x, double y, double x_size, double y_size)
{
    wall_type *w = walls;
    
    walls = (wall_type *) malloc(sizeof(wall_type));

    walls->x = x;
    walls->y = y;

    walls->x_size = x_size;
    walls->y_size = y_size;
       
    walls->next = w;
    
    num_walls++;
}
/* End of world_add_wall */



/* Free wall memory */
void world_class::free_walls( void )
{
    wall_type *tmp;
        
    while (walls != NULL) 
    {
        tmp = walls->next;

        free(walls);

        walls = tmp; // Unlink
    }

    num_walls = 0;
}    
/* End of world_free walls */


    
/* Draw walls to DC */
void world_class::draw_walls( HDC dc )
{
    wall_type *tmp = walls;

    SelectObject(dc, (HBRUSH)g_main->color1_brush);
    SelectObject(dc, (HPEN)g_main->color1_pen);

    while (tmp != NULL) 
    {
        RECT r;
        
        Rectangle(dc, get_screen_x(tmp->x),
                      get_screen_y(tmp->y),
                      get_screen_x(tmp->x + tmp->x_size),
                      get_screen_y(tmp->y + tmp->y_size));

        tmp = tmp->next;
    }
}    
/* End of world_draw_walls */


    

/* Add marker to the world */
void world_class::add_marker(double x, double y)
{
//    return;
    if (!show_markers) return;

    marker_type *m = markers;
    
    markers = (marker_type *) malloc(sizeof(marker_type));

    markers->x = x;
    markers->y = y;
       
    markers->next = m;
}
/* End of world_add_marker */



/* Free marker memory */
void world_class::free_markers( void )
{
    marker_type *tmp;
        
    while (markers != NULL) 
    {
        tmp = markers->next;


        free(markers);

        markers = tmp; // Unlink
    }
}    
/* End of world free markers */


/* draw markers to DC */
void world_class::draw_markers( HDC dc )
{
    marker_type *m = markers;

    if (!show_markers) return;

//return;

    double mx_size = 0.1;
    double my_size = 0.1;
    

    SelectObject(dc, (HBRUSH)g_main->blue_brush);

    int count = 500;

    while (m != NULL && count--) 
    {
    
        RECT r;
        
        get_bounding_rect(m->x,m->y,mx_size, my_size, r);

        Ellipse(dc, r.left, r.top, r.right, r.bottom);
        

        m = m->next;

    }
}    
/* End of world_draw_markers */

    
/* Add impact to world */
void world_class::add_impact(double x, double y)
{
    impact_type *i = impacts;
 
    while (i != NULL) 
    {
        double dist = (x-i->x)*(x-i->x) + (y-i->y)*(y-i->y);
        
        if (dist < 0.01) return;
        
        i = i->next;

    }

    i = impacts;
    
    impacts = (impact_type *) malloc(sizeof(impact_type));

    impacts->x = x;
    impacts->y = y;
       
    impacts->next = i;
    
    num_impact++;
    
    for (int i = 0; i < num_robots; i++)
        robots[i]->world_compute_map_grid();
}
/* End of world_add_impact */


/* Free impact memory */
void world_class::free_impacts( void )
{
    impact_type *tmp;
        
    while (impacts != NULL) 
    {
        tmp = impacts->next;

        free(impacts);

        impacts = tmp; // Unlink
    }
    num_impact = 0;
}    
/* End of world_free_impacts */


/* Draw imapcts to DC */
void world_class::draw_impacts( HDC dc )
{
    impact_type *i = impacts;

//return;
    double x_size = 0.2;
    double y_size = 0.2;
    

    SelectObject(dc, (HBRUSH)g_main->red_brush);
    SelectObject(dc, (HPEN)g_main->black_pen);
    

    while (i != NULL) 
    {
        RECT r;
        
        get_bounding_rect(i->x,i->y,x_size, y_size, r);

        Ellipse(dc, r.left, r.top, r.right, r.bottom);

        i = i->next;
    }
}    
/* End of world_draw_impacts */




void world_class::draw_target(HDC dc, double x, double y)
{
    if (target_enabled)
    {
        double target_size = 0.15;

        SelectObject(dc, g_main->green_brush);        

        Ellipse(dc, get_screen_x(x - target_size/2), 
                    get_screen_y(y - target_size/2), 
                    get_screen_x(x + target_size/2), 
                    get_screen_y(y + target_size/2));
                
    }    
}








/* Set target location */
void world_class::set_target(double x, double y)
{
    target_enabled = true;

    target_x = x;
    target_y = y;    

    for (int i = 0; i < num_robots; i++)
        robots[i]->world_compute_map_grid();
}
/* End of world_set_target */



/* Compute vector to closest impact */
void world_class::compute_impact_vector(double x, double y, point &v)
{
    v.x = 0;
    v.y = 0;
    
     if (num_impact == 0) return;
    
    impact_type *i = impacts;
   
    double closest_impact_x;
    double closest_impact_y;
   
    double closest_impact_dist = x_size * y_size;
   
   
    // Find closest impact 
    while (i != NULL) 
    {
        double x_diff = x - i->x;
        double y_diff = y - i->y;    
          
        double dist_squared = x_diff * x_diff + y_diff * y_diff;

        if (dist_squared < closest_impact_dist)
        {
            closest_impact_dist = dist_squared;
            closest_impact_x = i->x;
            closest_impact_y = i->y;            
        }
            
        i = i->next;        
     }


    v.x = x - closest_impact_x;
    v.y = y - closest_impact_y;        
}
/* End of compute_impact_vector */



/* Compute gradient at a point */
void world_class::compute_grad_point(robot_class *robot, double x, double y, double &dir, double &mag)
{
    double x_sum = 0, y_sum = 0;
    mag = 0;
    
    double dist;

    point v;

    // Compute vector to closest impact
    compute_impact_vector(x, y, v);
    
    dist = vect_mag(v.x, v.y);

    if (dist > 0)
    {
        double  force_x =  scale1 * v.x / pow(dist, 1.5);
        double  force_y =  scale1 * v.y / pow(dist, 1.5);

        x_sum += force_x;
        y_sum += force_y;

        mag += vect_mag(force_x, force_y);
    
    }
    
    
   // for (int i = 0; i < num_robots; i++)
    robot->compute_waypoint_vector(x, y, v);
    
    
    dist = vect_mag(v.x, v.y);
    
    if (dist > 0)
    {
        double force_x = scale2 * v.x;
        double force_y = scale2 * v.y;
        
       
        x_sum -= force_x;
        y_sum -= force_y;
    
        mag -= vect_mag(force_x, force_y);
    }
    

    
        robot->compute_map_nav( v, x, y);

    dist = vect_mag(v.x, v.y);
    
    if (dist > 0)
    {
        double force_x = scale3 * v.x/dist;
        double force_y = scale3 * v.y/dist;
        
        x_sum += force_x;
        y_sum += force_y;
    
        mag += vect_mag(force_x, force_y);
    }
    
    dir = -atan2(y_sum, x_sum);//+f_rand(-0.5, 0.5); 
}
/* End of compute_grad_point */



/* Draw gradient field */
void world_class::draw_grad(HDC dc)
{
    SelectObject(dc, g_main->green_pen);

    double x, y;
    
    int t_x_size = g_main->get_display_x() /NUM_X_GRAD;
    int t_y_size = g_main->get_display_y() /NUM_Y_GRAD;
    
    
    
//        log(MODULE_NAME "draw_grad x:   %d  y: %d\n", g_main->get_display_x(), g_main->get_display_y());

    
    
    HBRUSH br;
    HPEN pen;
    
    for (int xg = 0; xg < NUM_X_GRAD; xg++)
        for (int yg = 0; yg < NUM_Y_GRAD; yg++)
        {
            x = xg / (double)NUM_X_GRAD * x_size;
            y = yg / (double)NUM_Y_GRAD * y_size;

            double dir, mag;
            dir = 0; mag= 0;
             // compute_grad_point(x,y, dir, mag);


            double val = mag;
            
            if (val < -1) val = -1;
            if (val > 1) val = 1;
            
            int color = (int)(fabs(val) * 255);
            
            if (val < 0)
            {
                br = CreateSolidBrush(RGB(color, 0, 0));
                pen = CreatePen(PS_SOLID, 1, RGB(color,0,0));                
            }
            else
            { 
                br = CreateSolidBrush(RGB(0, 0, color));
                pen = CreatePen(PS_SOLID, 1, RGB(0,0,color));
            }


             SelectObject(dc, br);
             SelectObject(dc, pen);

             Rectangle(dc, get_screen_x(x)-t_x_size/2-1,
                           get_screen_y(y)-t_y_size/2-1,
                           get_screen_x(x)+t_x_size/2+1,
                           get_screen_y(y)+t_y_size/2+1);

             DeleteObject(br);
             DeleteObject(pen);

             SelectObject(dc, g_main->green_pen);    

           // double mag =  fabs(grad_mag[yg][xg]) / 100.0;
            
            mag = fabs(mag / 5.0);

            if (mag > 0.1) mag = 0.1;

            
         //   mag = 0.05;

            draw_arrow(dc, x, y, dir, mag);
    }
}
/* End of void world_draw_grad(HDC dc) */


/* Load world file */
void world_class::load( void )
{
    OPENFILENAME ofname;

    char filename[300] = {0};
             
    char filter[] = "World files\0*.dat\0\0";
           
    memset(&ofname, 0, sizeof(ofname));
             
    ofname.lStructSize = sizeof(ofname);
    ofname.lpstrFilter = filter; 
    ofname.lpstrFile = filename; 
    ofname.nMaxFile = 300; 
    ofname.lpstrFileTitle = NULL; 
//    ofname.nMaxFileTitle = 300; 
    ofname.lpstrInitialDir = "."; 

    // Get filename from user
    if (!GetOpenFileName(&ofname))
        return;

    FILE *f = fopen(filename, "rt");

    unsigned int num_walls;



    free_walls();
    
    double dummy;
    

/*
    fscanf(f, "%lf\n", &dummy);
    robot->set_x(dummy);
    
    fscanf(f, "%lf\n", &dummy);
    robot->set_y(dummy);
    
    fscanf(f, "%lf\n", &dummy);
    robot->set_dir(dummy);*/

    fscanf(f, "%d\n", &num_walls);
  
    for (int i = 0; i < num_walls && !feof(f); i++)
    {
        double x, y, x_size, y_size;
        
        fscanf(f, "%lf %lf %lf %lf\n", &x, &y, &x_size, &y_size);

        add_wall(x, y, x_size, y_size);
    }
}
/* End of world_load */


/* Save world data */
void world_class::save( void )
{
    OPENFILENAME ofname;
    char filename[300] = {0};
    char filter[] = "Map files\0*.dat\0\0";             
             
    memset(&ofname, 0, sizeof(ofname));
             
    ofname.lStructSize = sizeof(ofname);
    ofname.lpstrFilter = filter; 
    ofname.lpstrFile = filename; 
    ofname.nMaxFile = 300; 
    ofname.lpstrFileTitle = NULL; 
    ofname.lpstrInitialDir = "."; 

    // Get filename from user
    if (!GetSaveFileName(&ofname))
    {
        //write_console("erer");
        return;
    }


    FILE *f = fopen(filename, "wt");
    

  /*  fprintf(f, "%f\n", robot->get_x());
    fprintf(f, "%f\n", robot->get_y());
    fprintf(f, "%f\n", robot->get_dir());
    fprintf(f, "%d\n", num_walls);*/
    
    
    wall_type *w = walls;
    
    while(w != NULL)
    {
        fprintf(f, "%f %f %f %f\n", w->x, w->y, w->x_size, w->y_size);   
        
        w = w -> next;
    }
    
    fclose(f);
}
/* End of save world */




  void  world_class::random_target( void )
  {
  
      set_target(f_rand(1, x_size-1), f_rand(1, y_size-1));
  
  }









/* Check the space that the sensor tip occupies, and set values to match */
double world_class::check_vision(double x, double y, double dir, bool &found, point &p) 
{
    // Determine location of sensor tip
    double max_size = get_x_size() * get_y_size();

    line main_line;
    
    main_line.x1 = x;
    main_line.y1 = y;
    
    main_line.x2 = x + max_size * cos(dir);
    main_line.y2 = y - max_size * sin(dir);
        
    wall_type *w = walls;    
   
    found = false;
   
   
   
   
    point points[num_walls +num_robots];
    double dist[num_walls +num_robots];

    
    point intersect;
    line box;    
    
   
 
    
     unsigned int pts= 0;
    
    
    
    // Check against all walls
    while (w != NULL)
    {
        // The wall is checked using a line point to opposing corners
        box.x1 = w->x;
        box.y1 = w->y;
        
        box.x2 = w->x + w->x_size;
        box.y2 = w->y + w->y_size;
            
            double d = check_box_intersect(this, main_line, box, intersect);


//        dist[i] = check_box_intersect(this, main_line, box, intersect);
            
       // if (dist[i] > MAX_IR_DIST || dist[i] == -1)
//        dist[i] = MAX_IR_DIST;
            
        // check Intersection
        if  (d > 0 && d  < MAX_IR_DIST)
        {
            points[pts] = intersect;
            dist[pts] = d;

            p = intersect;
            found = true;  
            
                      pts++;          
        }

        w=w->next;
        
        //i++;
    }
    
/*
    
    for (int i = 0; i< num_robots; i++)
    {
        double s = (robots[i]->get_x_size() +robots[i]->get_y_size() ) / 2.0;
    
        double tx = x +  s * cos(dir);
        double ty = y -  s * sin(dir);
    
    

        box.x1 = robots[i]->get_x() - robots[i]->get_x_size();
        box.y1 = robots[i]->get_y() - robots[i]->get_y_size();

        box.x2 = robots[i]->get_x() + robots[i]->get_x_size();
        box.y2 = robots[i]->get_y() + robots[i]->get_y_size();

                    
         double d = check_box_intersect(this, main_line, box, intersect);


//        dist[i] = check_box_intersect(this, main_line, box, intersect);
            
       // if (dist[i] > MAX_IR_DIST || dist[i] == -1)
//        dist[i] = MAX_IR_DIST;
            
        // check Intersection
        if  (d > 0 && d  < MAX_IR_DIST)
        {
            points[pts] = intersect;
            dist[pts] = d;

            p = intersect;
            found = true;  
            
                      pts++;          
        }
    
    }
       
    */
    
    
    
    
    
    
      // Calculate and return distance */
    double smallest_dist = x_size + y_size;
    int smallest_index = -1;
    
    for (int i = 0; i < pts; i++)
    {
        if (dist[i] > 0 && dist[i] <= smallest_dist)
        {
            smallest_dist = dist[i];
            smallest_index = i;
        }
    }
  
  
  if (smallest_index == -1) return(MAX_IR_DIST);
  
    p = points[smallest_index];

         //   world_add_impact(points[smallest_index].x, points[smallest_index].y);    
    
    return(dist[smallest_index]);

}
/* End of check sensor */


double world_class::target_dist(double xv, double yv)
{
    double x_diff = xv - target_x;
    double y_diff = yv - target_y;

    return (x_diff * x_diff + y_diff * y_diff);
}





/* Check the space that the sensor tip occupies, and set values to match */
bool world_class::check_sensor(double x, double y, double dir, double dist) 
{
    // Determine location of sensor tip
    double sx = x + dist * cos(dir);
    double sy = y - dist * sin(dir);
    
    if (check_wall_point(sx, sy)) 
    {
        //world_add_impact(sx, sy);
        return(true);        
    }

    if (sx < 0 || sx > get_x_size()) 
    {
      //  world_add_impact(sx, sy);
        return(true);
    }

    if (sy < 0 || sy > get_y_size())
    { 
      // world_add_impact(sx, sy);
        return(true);
    }
    
    return(false);
}
/* End of check sensor */







